# import copy

# import dspy

# from .predict import Predict


# class Retry(Predict):
#     def __init__(self, module):
#         super().__init__(module.signature)
#         self.module = module
#         self.original_signature = module.signature
#         self.original_forward = module.forward
#         self.new_signature = self._create_new_signature(self.original_signature)

#     def _create_new_signature(self, signature):
#         # Add "Past" input fields for each output field
#         for key, value in signature.output_fields.items():
#             actual_prefix = value.json_schema_extra["prefix"].split(":")[0] + ":"
#             signature = signature.append(f"past_{key}", dspy.InputField(
#                 prefix="Previous " + actual_prefix,
#                 desc=f"past {actual_prefix[:-1]} with errors",
#                 format=value.json_schema_extra.get("format"),
#             ))

#         signature = signature.append("feedback", dspy.InputField(
#             prefix="Instructions:",
#             desc="Some instructions you must satisfy",
#             format=str,
#         ))

#         return signature

#     def forward(self, *, past_outputs, **kwargs):
#         # Take into account the possible new signature, as in TypedPredictor
#         new_signature = kwargs.pop("new_signature", None)
#         if new_signature:
#             self.original_signature = new_signature
#             self.new_signature = self._create_new_signature(self.original_signature)

#         # Convert the dict past_outputs={"answer": ...} to kwargs
#         # {past_answer=..., ...}
#         for key, value in past_outputs.items():
#             past_key = f"past_{key}"
#             if past_key in self.new_signature.input_fields:
#                 kwargs[past_key] = value
#         # Tell the wrapped module to use the new signature.
#         # Note: This only works if the wrapped module is a Predict or ChainOfThought.
#         kwargs["new_signature"] = self.new_signature
#         return self.original_forward(**kwargs)

#     def __call__(self, **kwargs):
#         copy.deepcopy(kwargs)
#         kwargs["_trace"] = False
#         kwargs.setdefault("demos", self.demos if self.demos is not None else [])

#         # perform backtracking
#         if dspy.settings.backtrack_to == self:
#             for key, value in dspy.settings.backtrack_to_args.items():
#                 kwargs.setdefault(key, value)
#             pred = self.forward(**kwargs)
#         else:
#             pred = self.module(**kwargs)

#         # now pop multiple reserved keys
#         # NOTE(shangyin) past_outputs seems not useful to include in demos,
#         # therefore dropped
#         for key in ["_trace", "demos", "signature", "new_signature", "config", "lm", "past_outputs"]:
#             kwargs.pop(key, None)

#         if dspy.settings.trace is not None:
#             trace = dspy.settings.trace
#             trace.append((self, {**kwargs}, pred))
#         return pred



# import math
# import random
# import sys
# import textwrap
# from collections import defaultdict
# from typing import Any

# import optuna

# import dspy
# from dspy.evaluate.evaluate import Evaluate
# from dspy.signatures import Signature
# from dspy.signatures.signature import signature_to_template
# from dspy.teleprompt import BootstrapFewShot
# from dspy.teleprompt.teleprompt import Teleprompter

# """
# USAGE SUGGESTIONS:

# The following code can be used to compile a optimized signature teleprompter using MIPRO, and evaluate it on an end task:

# ``` python
# from dspy.teleprompt import MIPRO

# teleprompter = MIPRO(prompt_model=prompt_model, task_model=task_model, metric=metric, num_candidates=10, init_temperature=1.0)
# kwargs = dict(num_threads=NUM_THREADS, display_progress=True, display_table=0)
# compiled_prompt_opt = teleprompter.compile(program, trainset=trainset[:TRAIN_NUM], num_trials=100, max_bootstrapped_demos=3, max_labeled_demos=5, eval_kwargs=kwargs)
# eval_score = evaluate(compiled_prompt_opt, devset=evalset[:EVAL_NUM], **kwargs)
# ```

# Note that this teleprompter takes in the following parameters:

# * prompt_model: The model used for prompt generation. When unspecified, defaults to the model set in settings (i.e., dspy.settings.configure(lm=task_model)).
# * task_model: The model used for running your task. When unspecified, defaults to the model set in settings (i.e., dspy.settings.configure(lm=task_model)).
# * metric: The task metric used for optimization.
# * num_candidates: The number of new prompts and sets of fewshot examples to generate and evaluate. Default=10.
# * init_temperature: The temperature used to generate new prompts. Higher roughly equals more creative. Default=1.0.
# * verbose: Tells the method whether or not to print intermediate steps.
# * track_stats: Tells the method whether or not to track statistics about the optimization process.
#                 If True, the method will track a dictionary with a key corresponding to the trial number, 
#                 and a value containing a dict with the following keys:
#                     * program: the program being evaluated at a given trial
#                     * score: the last average evaluated score for the program
#                     * pruned: whether or not this program was pruned
#                 This information will be returned as attributes of the best program.
# """


# class BasicGenerateInstruction(Signature):
#     """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English. Your task is to propose an instruction that will lead a good language model to perform the task well. Don't be afraid to be creative."""

#     basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
#     proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
#     proposed_prefix_for_output_field = dspy.OutputField(
#         desc="The string at the end of the prompt, which will help the model start solving the task",
#     )


# class BasicGenerateInstructionWithDataObservations(Signature):
#     """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English.  I will also give you some ``observations`` I have made about the dataset and task. Your task is to propose an instruction that will lead a good language model to perform the task well. Don't be afraid to be creative."""

#     basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
#     observations = dspy.InputField(desc="Observations about the dataset and task")
#     proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
#     proposed_prefix_for_output_field = dspy.OutputField(
#         desc="The string at the end of the prompt, which will help the model start solving the task",
#     )


# class BasicGenerateInstructionWithExamples(dspy.Signature):
#     """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English. Specifically, I will also provide you with the current ``basic instruction`` that is being used for this task. I will also provide you with some ``examples`` of the expected inputs and outputs.

#     Your task is to propose an instruction that will lead a good language model to perform the task well. Don't be afraid to be creative."""

#     # attempted_instructions = dspy.InputField(format=str, desc="Previously attempted task instructions, along with their resulting validation score, and an example of the instruction in use on a sample from our dataset.")
#     basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
#     # examples = dspy.InputField(desc="Example(s) of the task")
#     examples = dspy.InputField(desc="Example(s) of the task")
#     proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
#     proposed_prefix_for_output_field = dspy.OutputField(
#         desc="The string at the end of the prompt, which will help the model start solving the task",
#     )


# class BasicGenerateInstructionWithExamplesAndDataObservations(dspy.Signature):
#     """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English. Specifically, I will give you some ``observations`` I have made about the dataset and task, along with some ``examples`` of the expected inputs and outputs. I will also provide you with the current ``basic instruction`` that is being used for this task.

#     Your task is to propose a new improved instruction and prefix for the output field that will lead a good language model to perform the task well. Don't be afraid to be creative."""

#     observations = dspy.InputField(desc="Observations about the dataset and task")
#     examples = dspy.InputField(desc="Example(s) of the task")
#     basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
#     proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
#     proposed_prefix_for_output_field = dspy.OutputField(
#         desc="The string at the end of the prompt, which will help the model start solving the task",
#     )


# class ObservationSummarizer(dspy.Signature):
#     """Given a series of observations I have made about my dataset, please summarize them into a brief 2-3 sentence summary which highlights only the most important details."""

#     observations = dspy.InputField(desc="Observations I have made about my dataset")
#     summary = dspy.OutputField(
#         desc="Two to Three sentence summary of only the most significant highlights of my observations",
#     )


# class DatasetDescriptor(dspy.Signature):
#     (
#         """Given several examples from a dataset please write observations about trends that hold for most or all of the samples. """
#         """Some areas you may consider in your observations: topics, content, syntax, conciceness, etc. """
#         """It will be useful to make an educated guess as to the nature of the task this dataset will enable. Don't be afraid to be creative"""
#     )

#     examples = dspy.InputField(desc="Sample data points from the dataset")
#     observations = dspy.OutputField(desc="Somethings that holds true for most or all of the data you observed")


# class DatasetDescriptorWithPriorObservations(dspy.Signature):
#     (
#         """Given several examples from a dataset please write observations about trends that hold for most or all of the samples. """
#         """I will also provide you with a few observations I have already made.  Please add your own observations or if you feel the observations are comprehensive say 'COMPLETE' """
#         """Some areas you may consider in your observations: topics, content, syntax, conciceness, etc. """
#         """It will be useful to make an educated guess as to the nature of the task this dataset will enable. Don't be afraid to be creative"""
#     )

#     examples = dspy.InputField(desc="Sample data points from the dataset")
#     prior_observations = dspy.InputField(desc="Some prior observations I made about the data")
#     observations = dspy.OutputField(
#         desc="Somethings that holds true for most or all of the data you observed or COMPLETE if you have nothing to add",
#     )


# class MIPRO(Teleprompter):
#     def __init__(
#         self,
#         metric,
#         prompt_model=None,
#         task_model=None,
#         teacher_settings={},
#         num_candidates=10,
#         init_temperature=1.0,
#         verbose=False,
#         track_stats=True,
#         view_data_batch_size=10,
#     ):
#         self.num_candidates = num_candidates
#         self.metric = metric
#         self.init_temperature = init_temperature
#         self.prompt_model = prompt_model if prompt_model is not None else dspy.settings.lm
#         self.task_model = task_model if task_model is not None else dspy.settings.lm
#         self.verbose = verbose
#         self.track_stats = track_stats
#         self.teacher_settings = teacher_settings
#         self.view_data_batch_size = view_data_batch_size

#     def _print_full_program(self, program):
#         for i, predictor in enumerate(program.predictors()):
#             if self.verbose:
#                 print(f"Predictor {i}")
#             if self.verbose:
#                 print(f"i: {self._get_signature(predictor).instructions}")
#             *_, last_field = self._get_signature(predictor).fields.values()
#             if self.verbose:
#                 print(f"p: {last_field.json_schema_extra['prefix']}")
#             if self.verbose:
#                 print("\n")

#     def _print_model_history(self, model, n=1):
#         if self.verbose:
#             print(f"Model ({model}) History:")
#         model.inspect_history(n=n)

#     def _observe_data(self, trainset, max_iterations=10):
#         upper_lim = min(len(trainset), self.view_data_batch_size)
#         observation = dspy.Predict(DatasetDescriptor, n=1, temperature=1.0)(examples=(trainset[0:upper_lim].__repr__()))
#         observations = observation["observations"]

#         skips = 0
#         iterations = 0
#         for b in range(self.view_data_batch_size, len(trainset), self.view_data_batch_size):
#             upper_lim = min(len(trainset), b + self.view_data_batch_size)
#             output = dspy.Predict(DatasetDescriptorWithPriorObservations, n=1, temperature=1.0)(
#                 prior_observations=observations,
#                 examples=(trainset[b:upper_lim].__repr__()),
#             )
#             iterations += 1
#             if len(output["observations"]) >= 8 and output["observations"][:8].upper() == "COMPLETE":
#                 skips += 1
#                 if skips >= 5:
#                     break
#                 continue
#             if iterations >= max_iterations:
#                 break
#             observations += output["observations"]

#         summary = dspy.Predict(ObservationSummarizer, n=1, temperature=1.0)(observations=observations)

#         return summary.summary

#     def _create_example_string(self, fields, example):
#         # Building the output string
#         output = []
#         for field in fields:
#             name = field.name
#             separator = field.separator
#             input_variable = field.input_variable

#             # Determine the value from input_data or prediction_data
#             value = example.get(input_variable)

#             # Construct the string for the current field
#             field_str = f"{name}{separator}{value}"
#             output.append(field_str)

#         # Joining all the field strings
#         return "\n".join(output)

#     def _get_signature(self, predictor):
#         if hasattr(predictor, "extended_signature"):
#             return predictor.extended_signature
#         elif hasattr(predictor, "signature"):
#             return predictor.signature
#         return None

#     def _set_signature(self, predictor, updated_signature):
#         if hasattr(predictor, "extended_signature"):
#             predictor.extended_signature = updated_signature
#         elif hasattr(predictor, "signature"):
#             predictor.signature = updated_signature

#     def _generate_first_N_candidates(  # noqa: N802
#         self,
#         module: dspy.Module,
#         N: int,  # noqa: N803
#         view_data: bool,
#         view_examples: bool,
#         demo_candidates: dict,
#         devset,
#     ) -> tuple[dict, dict]:
#         candidates = {}
#         evaluated_candidates = defaultdict(dict)

#         if view_data:
#             # Create data observations
#             self.observations = None
#             with dspy.settings.context(lm=self.prompt_model):
#                 self.observations = self._observe_data(devset).replace("Observations:", "").replace("Summary:", "")

#         if view_examples:
#             example_sets = {}
#             for predictor in module.predictors():
#                 # Get all augmented examples
#                 example_set = {}
#                 all_sets_of_examples = demo_candidates[id(predictor)]  # Get all generated sets of examples
#                 for example_set_i, set_of_examples in enumerate(all_sets_of_examples):
#                     if example_set_i != 0:  # Skip the no examples case
#                         for example in set_of_examples:  # Get each individual example in the set
#                             if "augmented" in example and example["augmented"]:
#                                 if example_set_i not in example_set:
#                                     example_set[example_set_i] = []
#                                 fields_to_use = signature_to_template(predictor.signature).fields
#                                 _input_variable_names = list(self._get_signature(predictor).input_fields.keys())
#                                 example_string = self._create_example_string(fields_to_use, example)
#                                 example_set[example_set_i].append(example_string)
#                         example_sets[id(predictor)] = example_set
#                     else:
#                         example_set[example_set_i] = []
#                         example_sets[id(predictor)] = example_set

#         # Seed the prompt optimizer zero shot with just the instruction, generate BREADTH new prompts
#         for predictor in module.predictors():
#             basic_instruction = None
#             basic_prefix = None
#             basic_instruction = self._get_signature(predictor).instructions
#             *_, last_field = self._get_signature(predictor).fields.values()
#             basic_prefix = last_field.json_schema_extra["prefix"]
#             with dspy.settings.context(lm=self.prompt_model):
#                 # Data & Examples
#                 if view_data and view_examples:
#                     if 1 not in example_sets[id(predictor)]:
#                         raise ValueError("No examples found for the given predictor")
#                     instruct = None
#                     for i in range(1, self.num_candidates):
#                         new_instruct = dspy.Predict(
#                             BasicGenerateInstructionWithExamplesAndDataObservations,
#                             n=1,
#                             temperature=self.init_temperature,
#                         )(
#                             basic_instruction=basic_instruction,
#                             observations=self.observations,
#                             examples=example_sets[id(predictor)][i],
#                         )
#                         if not instruct:
#                             instruct = new_instruct
#                         else:
#                             instruct.completions.proposed_instruction.extend(
#                                 new_instruct.completions.proposed_instruction,
#                             )
#                             instruct.completions.proposed_prefix_for_output_field.extend(
#                                 new_instruct.completions.proposed_prefix_for_output_field,
#                             )
#                 # Just data
#                 elif view_data:
#                     instruct = dspy.Predict(
#                         BasicGenerateInstructionWithDataObservations,
#                         n=N - 1,
#                         temperature=self.init_temperature,
#                     )(basic_instruction=basic_instruction, observations=self.observations)
#                 # Just examples
#                 elif view_examples:
#                     instruct = None
#                     for i in range(1, self.num_candidates):  # Note: skip over the first example set which is empty
#                         new_instruct = dspy.Predict(
#                             BasicGenerateInstructionWithExamples,
#                             n=1,
#                             temperature=self.init_temperature,
#                         )(
#                             basic_instruction=basic_instruction,
#                             examples=example_sets[id(predictor)][i],
#                         )
#                         if not instruct:
#                             instruct = new_instruct
#                         else:
#                             instruct.completions.proposed_instruction.extend(
#                                 new_instruct.completions.proposed_instruction,
#                             )
#                             instruct.completions.proposed_prefix_for_output_field.extend(
#                                 new_instruct.completions.proposed_prefix_for_output_field,
#                             )
#                 # Neither
#                 else:
#                     instruct = dspy.Predict(BasicGenerateInstruction, n=N - 1, temperature=self.init_temperature)(
#                         basic_instruction=basic_instruction,
#                     )

#             # Add in our initial prompt as a candidate as well
#             instruct.completions.proposed_instruction.insert(0, basic_instruction)
#             instruct.completions.proposed_prefix_for_output_field.insert(0, basic_prefix)
#             candidates[id(predictor)] = instruct.completions
#             evaluated_candidates[id(predictor)] = {}

#         if self.verbose:
#             self._print_model_history(self.prompt_model)

#         return candidates, evaluated_candidates

#     def compile(
#         self,
#         student: dspy.Program,
#         *,
#         trainset: list[dspy.Example],
#         num_trials: int,
#         max_bootstrapped_demos: int,
#         max_labeled_demos: int,
#         eval_kwargs: dict[str, Any],
#         seed=42,
#         view_data=True,
#         view_examples=True,
#         requires_permission_to_run=True,
#     ) -> dspy.Program:
#         # Define ANSI escape codes for colors
#         RED = "\033[91m"
#         YELLOW = "\033[93m"
#         BLUE = "\033[94m"
#         BOLD = "\033[1m"
#         ENDC = "\033[0m"  # Resets the color to default

#         random.seed(seed)

#         estimated_task_model_calls_wo_module_calls = len(trainset) * num_trials  # M * T * P
#         estimated_prompt_model_calls = 10 + self.num_candidates * len(
#             student.predictors(),
#         )  # num data summary calls + N * P

#         user_message = textwrap.dedent(f"""\
#             {YELLOW}{BOLD}WARNING: Projected Language Model (LM) Calls{ENDC}

#             Please be advised that based on the parameters you have set, the maximum number of LM calls is projected as follows:

#             {YELLOW}- Task Model: {BLUE}{BOLD}{len(trainset)}{ENDC}{YELLOW} examples in dev set * {BLUE}{BOLD}{num_trials}{ENDC}{YELLOW} trials * {BLUE}{BOLD}# of LM calls in your program{ENDC}{YELLOW} = ({BLUE}{BOLD}{estimated_task_model_calls_wo_module_calls} * # of LM calls in your program{ENDC}{YELLOW}) task model calls{ENDC}
#             {YELLOW}- Prompt Model: # data summarizer calls (max {BLUE}{BOLD}10{ENDC}{YELLOW}) + {BLUE}{BOLD}{self.num_candidates}{ENDC}{YELLOW} * {BLUE}{BOLD}{len(student.predictors())}{ENDC}{YELLOW} lm calls in program = {BLUE}{BOLD}{estimated_prompt_model_calls}{ENDC}{YELLOW} prompt model calls{ENDC}

#             {YELLOW}{BOLD}Estimated Cost Calculation:{ENDC}

#             {YELLOW}Total Cost = (Number of calls to task model * (Avg Input Token Length per Call * Task Model Price per Input Token + Avg Output Token Length per Call * Task Model Price per Output Token) 
#                         + (Number of calls to prompt model * (Avg Input Token Length per Call * Task Prompt Price per Input Token + Avg Output Token Length per Call * Prompt Model Price per Output Token).{ENDC}

#             For a preliminary estimate of potential costs, we recommend you perform your own calculations based on the task
#             and prompt models you intend to use. If the projected costs exceed your budget or expectations, you may consider:

#             {YELLOW}- Reducing the number of trials (`num_trials`), the size of the trainset, or the number of LM calls in your program.{ENDC}
#             {YELLOW}- Using a cheaper task model to optimize the prompt.{ENDC}""")

#         user_confirmation_message = textwrap.dedent(f"""\
#             To proceed with the execution of this program, please confirm by typing {BLUE}'y'{ENDC} for yes or {BLUE}'n'{ENDC} for no.

#             If you would like to bypass this confirmation step in future executions, set the {YELLOW}`requires_permission_to_run`{ENDC} flag to {YELLOW}`False` when calling compile.{ENDC}

#             {YELLOW}Awaiting your input...{ENDC}
#         """)

#         print(f"""{RED}{BOLD}WARNING: MIPRO has been deprecated and replaced with MIPROv2.  MIPRO will be removed in a future release. {ENDC}""")
#         print(user_message)

#         sys.stdout.flush()  # Flush the output buffer to force the message to print

#         run = True
#         if requires_permission_to_run:
#             print(user_confirmation_message)
#             user_input = input("Do you wish to continue? (y/n): ").strip().lower()
#             if user_input != "y":
#                 print("Compilation aborted by the user.")
#                 run = False

#         if run:
#             # Set up program and evaluation function
#             module = student.deepcopy()
#             evaluate = Evaluate(devset=trainset, metric=self.metric, **eval_kwargs)

#             # In the case where the bootstrapped and labeled demos are set to 0, we'll still bootstrap examples to use in our meta prompt
#             if (
#                 max_bootstrapped_demos == 0 and max_labeled_demos == 0
#             ):  # TODO: address case when max_bootstrapped alone is 0
#                 max_bootstrapped_demos_for_candidate_gen = 1
#                 max_labeled_demos_for_candidate_gen = 1  # TODO: this might only need to be 0
#             else:
#                 max_bootstrapped_demos_for_candidate_gen = max_bootstrapped_demos
#                 max_labeled_demos_for_candidate_gen = max_labeled_demos

#             # Generate N few shot example sets
#             demo_candidates = {}
#             for i in range(self.num_candidates):
#                 if i == 0:  # Story empty set of demos as default for index 0
#                     for module_p in module.predictors():
#                         if id(module_p) not in demo_candidates:
#                             demo_candidates[id(module_p)] = []
#                         demo_candidates[id(module_p)].append([])
#                 else:
#                     if self.verbose:
#                         print(f"Creating basic bootstrap: {i}/{self.num_candidates-1}")

#                     # Create a new basic bootstrap few - shot program .
#                     rng = random.Random(i)
#                     shuffled_trainset = trainset[:]  # Create a copy of devset
#                     rng.shuffle(shuffled_trainset)  # Shuffle the copy
#                     tp = BootstrapFewShot(
#                         metric=self.metric,
#                         max_bootstrapped_demos=max_bootstrapped_demos_for_candidate_gen,
#                         max_labeled_demos=max_labeled_demos_for_candidate_gen,
#                         teacher_settings=self.teacher_settings,
#                     )
#                     candidate_program = tp.compile(student=module.deepcopy(), trainset=shuffled_trainset)

#                     # Store the candidate demos
#                     for module_p, candidate_p in zip(module.predictors(), candidate_program.predictors()):
#                         if id(module_p) not in demo_candidates:
#                             demo_candidates[id(module_p)] = []
#                         demo_candidates[id(module_p)].append(candidate_p.demos)

#             # Generate N candidate prompts
#             instruction_candidates, _ = self._generate_first_N_candidates(
#                 module,
#                 self.num_candidates,
#                 view_data,
#                 view_examples,
#                 demo_candidates,
#                 trainset,
#             )

#             # Reset demo_candidates to None for our optimization if the user asked for no fewshot examples
#             if max_bootstrapped_demos == 0 and max_labeled_demos == 0:
#                 demo_candidates = None

#             # Initialize variables to store the best program and its score
#             best_score = float("-inf")
#             best_program = None
#             trial_num = 0

#             trial_logs = {}

#             # Define our trial objective
#             def create_objective(baseline_program, instruction_candidates, demo_candidates, evaluate, trainset):
#                 def objective(trial):
#                     nonlocal best_program, best_score, trial_num, trial_logs  # Allow access to the outer variables
#                     candidate_program = baseline_program.deepcopy()

#                     # Suggest the instruction to use for our predictor
#                     print(f"Starting trial #{trial_num}")
#                     trial_logs[trial_num] = {}

#                     for p_old, p_new in zip(baseline_program.predictors(), candidate_program.predictors()):
#                         # Get instruction candidates for our given predictor
#                         p_instruction_candidates = instruction_candidates[id(p_old)]
#                         if demo_candidates:
#                             p_demo_candidates = demo_candidates[id(p_old)]

#                         # Suggest the index of the instruction candidate to use in our trial
#                         instruction_idx = trial.suggest_categorical(
#                             f"{id(p_old)}_predictor_instruction",
#                             range(len(p_instruction_candidates)),
#                         )
#                         if demo_candidates:
#                             demos_idx = trial.suggest_categorical(
#                                 f"{id(p_old)}_predictor_demos",
#                                 range(len(p_demo_candidates)),
#                             )
#                         trial_logs[trial_num][f"{id(p_old)}_predictor_instruction"] = instruction_idx
#                         if demo_candidates:
#                             trial_logs[trial_num][f"{id(p_old)}_predictor_demos"] = demos_idx

#                         # Get the selected instruction candidate
#                         selected_candidate = p_instruction_candidates[instruction_idx]
#                         selected_instruction = selected_candidate.proposed_instruction.strip('"').strip()
#                         selected_prefix = selected_candidate.proposed_prefix_for_output_field.strip('"').strip()

#                         # Use this candidates in our program
#                         *_, last_field = self._get_signature(p_new).fields.keys()
#                         updated_signature = (
#                             self._get_signature(p_new)
#                             .with_instructions(selected_instruction)
#                             .with_updated_fields(last_field, prefix=selected_prefix)
#                         )
#                         self._set_signature(p_new, updated_signature)

#                         # Get the selected demos
#                         if demo_candidates:
#                             selected_demos = p_demo_candidates[demos_idx]

#                         # Use these demos in our program
#                         if demo_candidates:
#                             p_new.demos = selected_demos

#                     if self.verbose:
#                         print("Evaling the following program:")
#                     if self.verbose:
#                         self._print_full_program(candidate_program)
#                     trial_logs[trial_num]["program"] = candidate_program

#                     # Evaluate with the new prompts
#                     total_score = 0
#                     batch_size = 100
#                     num_batches = math.ceil(len(trainset) / batch_size)

#                     for i in range(num_batches):
#                         start_index = i * batch_size
#                         end_index = min((i + 1) * batch_size, len(trainset))
#                         split_trainset = trainset[start_index:end_index]
#                         split_score = evaluate(candidate_program, devset=split_trainset, display_table=0)
#                         if self.verbose:
#                             print(f"{i}st split score: {split_score}")

#                         total_score += split_score * len(split_trainset)
#                         curr_weighted_avg_score = total_score / min((i + 1) * 100, len(trainset))
#                         if self.verbose:
#                             print(f"curr average score: {curr_weighted_avg_score}")

#                         trial.report(curr_weighted_avg_score, i)

#                         # Handle pruning based on the intermediate value.
#                         if trial.should_prune():
#                             print("Trial pruned.")
#                             trial_logs[trial_num]["score"] = curr_weighted_avg_score
#                             trial_logs[trial_num]["pruned"] = True
#                             trial_num += 1
#                             raise optuna.TrialPruned()

#                     if self.verbose:
#                         print(f"Fully evaled score: {curr_weighted_avg_score}")
#                     if self.verbose:
#                         self._print_model_history(self.task_model, n=1)
#                     score = curr_weighted_avg_score

#                     trial_logs[trial_num]["score"] = curr_weighted_avg_score
#                     trial_logs[trial_num]["pruned"] = False

#                     # Update the best program if the current score is better
#                     if score > best_score:
#                         best_score = score
#                         best_program = candidate_program.deepcopy()

#                     trial_num += 1

#                     return score

#                 return objective

#             # Run the trial
#             objective_function = create_objective(module, instruction_candidates, demo_candidates, evaluate, trainset)
#             sampler = optuna.samplers.TPESampler(seed=seed)
#             study = optuna.create_study(direction="maximize", sampler=sampler)
#             _score = study.optimize(objective_function, n_trials=num_trials)

#             if best_program is not None and self.track_stats:
#                 best_program.trial_logs = trial_logs

#             print(f"Returning {best_program} from continue_program")
#             return best_program
#         return None



import re
import string
import unicodedata
from collections import Counter

from dspy.dsp.utils.utils import print_message


def EM(prediction, answers_list):
    assert isinstance(answers_list, list)

    return max(em_score(prediction, ans) for ans in answers_list)


def F1(prediction, answers_list):
    assert isinstance(answers_list, list)

    return max(f1_score(prediction, ans) for ans in answers_list)


def HotPotF1(prediction, answers_list):
    assert isinstance(answers_list, list)

    return max(hotpot_f1_score(prediction, ans) for ans in answers_list)


def normalize_text(s):
    s = unicodedata.normalize('NFD', s)

    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def em_score(prediction, ground_truth):
    return normalize_text(prediction) == normalize_text(ground_truth)


# See: https://github.com/hotpotqa/hotpot/blob/master/hotpot_evaluate_v1.py
# See: https://rajpurkar.github.io/SQuAD-explorer/ under Evaluation Script
# See: QReCC's

def f1_score(prediction, ground_truth):
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print_message(
            "\n#> F1 Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n")

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)

    return f1


def hotpot_f1_score(prediction, ground_truth):
    normalized_prediction = normalize_text(prediction)
    normalized_ground_truth = normalize_text(ground_truth)

    if normalized_prediction in ['yes', 'no', 'noanswer'] and normalized_prediction != normalized_ground_truth:
        return 0
    if normalized_ground_truth in ['yes', 'no', 'noanswer'] and normalized_prediction != normalized_ground_truth:
        return 0

    prediction_tokens = normalized_prediction.split()
    ground_truth_tokens = normalized_ground_truth.split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def precision_score(prediction, ground_truth):
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print_message(
            "\n#> F1 Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n")

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)

    return precision



import numpy as np
from typing import Any, List, Optional
from dspy.utils.unbatchify import Unbatchify

# TODO: Add .save and .load methods!


class Embeddings:
    def __init__(
        self,
        corpus: List[str],
        embedder,
        k: int = 5,
        callbacks: Optional[List[Any]] = None,
        cache: bool = False,
        brute_force_threshold: int = 20_000,
        normalize: bool = True
    ):
        assert cache is False, "Caching is not supported for embeddings-based retrievers"

        self.embedder = embedder
        self.k = k
        self.corpus = corpus
        self.normalize = normalize

        self.corpus_embeddings = self.embedder(self.corpus)
        self.corpus_embeddings = self._normalize(self.corpus_embeddings) if self.normalize else self.corpus_embeddings

        self.index = self._build_faiss() if len(corpus) >= brute_force_threshold else None
        self.search_fn = Unbatchify(self._batch_forward)

    def __call__(self, query: str):
        return self.forward(query)

    def forward(self, query: str):
        import dspy
        return dspy.Prediction(passages=self.search_fn(query))

    def _batch_forward(self, queries: List[str]):
        q_embeds = self.embedder(queries)
        q_embeds = self._normalize(q_embeds) if self.normalize else q_embeds

        pids = self._faiss_search(q_embeds, self.k * 10) if self.index else None
        pids = np.tile(np.arange(len(self.corpus)), (len(queries), 1)) if pids is None else pids
        
        return self._rerank_and_predict(q_embeds, pids)

    def _build_faiss(self):
        nbytes = 32
        partitions = int(2 * np.sqrt(len(self.corpus)))
        dim = self.corpus_embeddings.shape[1]

        try:
            import faiss
        except ImportError:
            raise ImportError("Please `pip install faiss-cpu` or increase `brute_force_threshold` to avoid FAISS.")

        quantizer = faiss.IndexFlatL2(dim)
        index = faiss.IndexIVFPQ(quantizer, dim, partitions, nbytes, 8)

        print(f"Training a {nbytes}-byte FAISS index with {partitions} partitions, based on "
              f"{len(self.corpus)} x {dim}-dim embeddings")
        index.train(self.corpus_embeddings)
        index.add(self.corpus_embeddings)
        index.nprobe = min(16, partitions)

        return index

    def _faiss_search(self, query_embeddings: np.ndarray, num_candidates: int):
        return self.index.search(query_embeddings, num_candidates)[1]

    def _rerank_and_predict(self, q_embeds: np.ndarray, candidate_indices: np.ndarray):
        candidate_embeddings = self.corpus_embeddings[candidate_indices]
        scores = np.einsum('qd,qkd->qk', q_embeds, candidate_embeddings)

        top_k_indices = np.argsort(-scores, axis=1)[:, :self.k]
        top_indices = candidate_indices[np.arange(len(q_embeds))[:, None], top_k_indices]

        return [[self.corpus[idx] for idx in indices] for indices in top_indices]

    def _normalize(self, embeddings: np.ndarray):
        norms = np.linalg.norm(embeddings, axis=1, keepdims=True)
        return embeddings / np.maximum(norms, 1e-10)



# import copy
# import random
# from functools import reduce

# from langchain_core.pydantic_v1 import Extra
# from langchain_core.runnables import Runnable

# import dspy.dsp as dsp
# import dspy
# from dspy.predict.parameter import Parameter
# from dspy.predict.predict import Predict
# from dspy.primitives.prediction import Prediction
# from dspy.signatures.field import OldInputField, OldOutputField
# from dspy.signatures.signature import infer_prefix

# # TODO: This class is currently hard to test, because it hardcodes gpt-4 usage:
# # gpt4T = dspy.OpenAI(model='gpt-4-1106-preview', max_tokens=4000, model_type='chat')

# class Template2Signature(dspy.Signature):
#     """You are a processor for prompts. I will give you a prompt template (Python f-string) for an arbitrary task for other LMs.
# Your job is to prepare three modular pieces: (i) any essential task instructions or guidelines, (ii) a list of variable names for inputs, (iv) the variable name for output."""

#     template = dspy.InputField(format=lambda x: f"```\n\n{x.strip()}\n\n```\n\nLet's now prepare three modular pieces.")
#     essential_instructions = dspy.OutputField()
#     input_keys = dspy.OutputField(desc='comma-separated list of valid variable names')
#     output_key = dspy.OutputField(desc='a valid variable name')


# class ShallowCopyOnly:
#     def __init__(self, obj): self.obj = obj
#     def __getattr__(self, item): return getattr(self.obj, item)
#     def __deepcopy__(self, memo): return ShallowCopyOnly(copy.copy(self.obj))


# class LangChainPredictMetaClass(type(Predict), type(Runnable)):
#     pass

# class LangChainPredict(Predict, Runnable, metaclass=LangChainPredictMetaClass):
#     class Config: extra = Extra.allow  # Allow extra attributes that are not defined in the model

#     def __init__(self, prompt, llm, **config):
#         Runnable.__init__(self)
#         Parameter.__init__(self)

#         self.langchain_llm = ShallowCopyOnly(llm)

#         try: langchain_template = '\n'.join([msg.prompt.template for msg in prompt.messages])
#         except AttributeError: langchain_template = prompt.template

#         self.stage = random.randbytes(8).hex()
#         self.signature, self.output_field_key = self._build_signature(langchain_template)
#         self.config = config
#         self.reset()

#     def reset(self):
#         self.lm = None
#         self.traces = []
#         self.train = []
#         self.demos = []

#     def dump_state(self, save_verbose=False):
#         """save_verbose is set as a default argument to support the inherited Parameter interface for dump_state"""
#         state_keys = ["lm", "traces", "train", "demos"]
#         return {k: getattr(self, k) for k in state_keys}

#     def load_state(self, state):
#         for name, value in state.items():
#             setattr(self, name, value)

#         self.demos = [dspy.Example(**x) for x in self.demos]
    
#     def __call__(self, *arg, **kwargs):
#         if len(arg) > 0: kwargs = {**arg[0], **kwargs}
#         return self.forward(**kwargs)
    
#     def _build_signature(self, template):
#         gpt4T = dspy.OpenAI(model='gpt-4-1106-preview', max_tokens=4000, model_type='chat')

#         with dspy.context(lm=gpt4T): parts = dspy.Predict(Template2Signature)(template=template)

#         inputs = {k.strip(): OldInputField() for k in parts.input_keys.split(',')}
#         outputs = {k.strip(): OldOutputField() for k in parts.output_key.split(',')}

#         for k, v in inputs.items():
#             v.finalize(k, infer_prefix(k))  # TODO: Generate from the template at dspy.Predict(Template2Signature)

#         for k, v in outputs.items():
#             output_field_key = k
#             v.finalize(k, infer_prefix(k))

#         return dsp.Template(parts.essential_instructions, **inputs, **outputs), output_field_key

#     def forward(self, **kwargs):
#         # Extract the three privileged keyword arguments.
#         signature = kwargs.pop("signature", self.signature)
#         demos = kwargs.pop("demos", self.demos)
#         config = dict(**self.config, **kwargs.pop("config", {}))

#         prompt = signature(dsp.Example(demos=demos, **kwargs))
#         output = self.langchain_llm.invoke(prompt, **config)

#         try: content = output.content
#         except AttributeError: content = output

#         pred = Prediction.from_completions([{self.output_field_key: content}], signature=signature)

#         # print('#> len(demos) =', len(demos))
#         # print(f"#> {prompt}")
#         # print(f"#> PRED = {content}\n\n\n")
#         dspy.settings.langchain_history.append((prompt, pred))
            
#         if dsp.settings.trace is not None:
#             trace = dsp.settings.trace
#             trace.append((self, {**kwargs}, pred))

#         return output
    
#     def invoke(self, d, *args, **kwargs):
#         # print(d)
#         return self.forward(**d)


# # Almost good but need output parsing for the fields!
# # TODO: Use template.extract(example, p)

# # class LangChainOfThought(LangChainPredict):
# #     def __init__(self, signature, **config):
# #         super().__init__(signature, **config)

# #         signature = self.signature
# #         *keys, last_key = signature.kwargs.keys()
# #         rationale_type = dsp.Type(prefix="Reasoning: Let's think step by step in order to",
# #                                   desc="${produce the " + last_key + "}. We ...")

# #         extended_kwargs = {key: signature.kwargs[key] for key in keys}
# #         extended_kwargs.update({"rationale": rationale_type, last_key: signature.kwargs[last_key]})
# #         self.extended_signature = dsp.Template(signature.instructions, **extended_kwargs)

# #     def forward(self, **kwargs):
# #         signature = self.extended_signature
# #         return super().forward(signature=signature, **kwargs)


# class LangChainModule(dspy.Module):
#     def __init__(self, lcel):
#         super().__init__()
        
#         modules = []
#         for name, node in lcel.get_graph().nodes.items():
#             if isinstance(node.data, LangChainPredict): modules.append(node.data)

#         self.modules = modules
#         self.chain = lcel
    
#     def forward(self, **kwargs):
#         output_keys = ['output', self.modules[-1].output_field_key]
#         output = self.chain.invoke(dict(**kwargs))
        
#         try: output = output.content
#         except Exception: pass

#         return dspy.Prediction({k: output for k in output_keys})
    
#     def invoke(self, d, *args, **kwargs):
#         return self.forward(**d).output

#     def reset_copy(self):
#         """Override `reset_copy()` to skip parts that cannot be deep copied."""
#         if self.chain is None:
#             obj = copy.deepcopy(self)
#         else:
#             chain = self.chain
#             steps = []
#             for step in chain.steps:
#                 if isinstance(step, LangChainPredict):
#                     # Only copy the LangChainPredict object.
#                     steps.append(copy.deepcopy(step))
#                 else:
#                     steps.append(step)

#             chain_copy = reduce(lambda x, y: x | y, steps)
#             # Temporarily remove the chain from `self` instance for deep copying.
#             self.chain = None
#             obj = copy.deepcopy(self)
#             obj.chain = chain_copy
#             # Put back the chain to `self` instance.
#             self.chain = chain

#         for param in obj.parameters():
#             param.reset()

#         return obj



import re
import time
from datetime import datetime
from typing import Any, Dict, List, Optional

import openai

from dspy.clients.provider import TrainingJob, Provider
from dspy.clients.utils_finetune import DataFormat, TrainingStatus, save_data


_OPENAI_MODELS = [
  'gpt-4-turbo',
  'gpt-4-turbo-2024-04-09',
  'tts-1',
  'tts-1-1106',
  'chatgpt-4o-latest',
  'dall-e-2',
  'whisper-1',
  'gpt-3.5-turbo-instruct',
  'gpt-3.5-turbo',
  'gpt-3.5-turbo-0125',
  'babbage-002',
  'davinci-002',
  'gpt-4o-mini-2024-07-18',
  'gpt-4o',
  'dall-e-3',
  'gpt-4o-mini',
  'gpt-4o-2024-08-06',
  'gpt-4o-2024-05-13',
  'o1-preview',
  'gpt-4o-audio-preview-2024-10-01',
  'o1-mini-2024-09-12',
  'gpt-4o-audio-preview',
  'tts-1-hd',
  'tts-1-hd-1106',
  'o1-preview-2024-09-12',
  'o1-mini',
  'gpt-4-1106-preview',
  'text-embedding-ada-002',
  'gpt-3.5-turbo-16k',
  'text-embedding-3-small',
  'text-embedding-3-large',
  'gpt-4o-realtime-preview-2024-10-01',
  'gpt-4o-realtime-preview',
  'gpt-3.5-turbo-1106',
  'gpt-4-0613',
  'gpt-4-turbo-preview',
  'gpt-4-0125-preview',
  'gpt-4',
  'gpt-3.5-turbo-instruct-0914'
]


class TrainingJobOpenAI(TrainingJob):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider_file_id = None
        self.provider_job_id = None

    def cancel(self):
        # Cancel the provider job
        if OpenAIProvider.does_job_exist(self.provider_job_id):
            status = self.status()
            if OpenAIProvider.is_terminal_training_status(status):
                err_msg = "Jobs that are complete cannot be canceled."
                err_msg += f" Job with ID {self.provider_job_id} is done."
                raise Exception(err_msg)
            openai.fine_tuning.jobs.cancel(self.provider_job_id)
            self.provider_job_id = None

        # Delete the provider file
        if self.provider_file_id is not None:
            if OpenAIProvider.does_file_exist(self.provider_file_id):
                openai.files.delete(self.provider_file_id)
            self.provider_file_id = None

        # Call the super's cancel method after the custom cancellation logic
        super().cancel()

    def status(self) -> TrainingStatus:
        status = OpenAIProvider.get_training_status(self.provider_job_id)
        return status


class OpenAIProvider(Provider):
    
    def __init__(self):
        super().__init__()
        self.finetunable = True
        self.TrainingJob = TrainingJobOpenAI

    @staticmethod
    def is_provider_model(model: str) -> bool:
        # Filter the provider_prefix, if exists
        provider_prefix = "openai/"
        if model.startswith(provider_prefix):
            model = model[len(provider_prefix):]

        # Check if the model is a base OpenAI model
        # TODO(enhance) The following list can be replaced with
        # openai.models.list(), but doing so might require a key. Is there a
        # way to get the list of models without a key?
        valid_model_names = _OPENAI_MODELS
        if model in valid_model_names:
            return True

        # Check if the model is a fine-tuned OpneAI model. Fine-tuned OpenAI
        # models have the prefix "ft:<BASE_MODEL_NAME>:", followed by a string
        # specifying the fine-tuned model. The following RegEx pattern is used
        # to match the base model name.
        # TODO(enhance): This part can be updated to match the actual fine-tuned
        # model names by making a call to the OpenAI API to be more exact, but
        # this might require an API key with the right permissions.
        match = re.match(r"ft:([^:]+):", model)
        if match and match.group(1) in valid_model_names:
            return True

        return False

    @staticmethod
    def finetune(
        job: TrainingJobOpenAI,
        model: str,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        data_format: Optional[DataFormat] = None,
    ) -> str:
        print("[OpenAI Provider] Validating the data format")
        OpenAIProvider.validate_data_format(data_format)

        print("[OpenAI Provider] Saving the data to a file")
        data_path = save_data(train_data)
        print(f"[OpenAI Provider] Data saved to {data_path}")

        print("[OpenAI Provider] Uploading the data to the provider")
        provider_file_id = OpenAIProvider.upload_data(data_path)
        job.provider_file_id = provider_file_id

        print("[OpenAI Provider] Starting remote training")
        provider_job_id = OpenAIProvider.start_remote_training(
            train_file_id=job.provider_file_id,
            model=model,
            train_kwargs=train_kwargs,
        )
        job.provider_job_id = provider_job_id
        print(f"[OpenAI Provider] Job started with the OpenAI Job ID {provider_job_id}")

        print("[OpenAI Provider] Waiting for training to complete")
        # TODO(feature): Could we stream OAI logs?
        OpenAIProvider.wait_for_job(job)

        print("[OpenAI Provider] Attempting to retrieve the trained model")
        model = OpenAIProvider.get_trained_model(job)
        print(f"[OpenAI Provider] Model retrieved: {model}")

        return model

    @staticmethod
    def does_job_exist(job_id: str) -> bool:
        try:
            # TODO(nit): This call may fail for other reasons. We should check
            # the error message to ensure that the job does not exist.
            openai.fine_tuning.jobs.retrieve(job_id)
            return True
        except Exception:
            return False

    @staticmethod
    def does_file_exist(file_id: str) -> bool:
        try:
            # TODO(nit): This call may fail for other reasons. We should check
            # the error message to ensure that the file does not exist.
            openai.files.retrieve(file_id)
            return True
        except Exception:
            return False


    @staticmethod
    def is_terminal_training_status(status: TrainingStatus) -> bool:
        return status in [
            TrainingStatus.succeeded,
            TrainingStatus.failed,
            TrainingStatus.cancelled,
        ]
    
    @staticmethod
    def get_training_status(job_id: str) -> TrainingStatus:
        provider_status_to_training_status = {
            "validating_files": TrainingStatus.pending,
            "queued": TrainingStatus.pending,
            "running": TrainingStatus.running,
            "succeeded": TrainingStatus.succeeded,
            "failed": TrainingStatus.failed,
            "cancelled": TrainingStatus.cancelled,
        }

        # Check if there is an active job
        if job_id is None:
            print("There is no active job.")
            return TrainingStatus.not_started

        err_msg = f"Job with ID {job_id} does not exist."
        assert OpenAIProvider.does_job_exist(job_id), err_msg

        # Retrieve the provider's job and report the status
        provider_job = openai.fine_tuning.jobs.retrieve(job_id)
        provider_status = provider_job.status
        status = provider_status_to_training_status[provider_status]

        return status

    @staticmethod
    def validate_data_format(data_format: DataFormat):
        supported_data_formats = [
            DataFormat.chat,
            DataFormat.completion,
        ]
        if data_format not in supported_data_formats:
            err_msg = f"OpenAI does not support the data format {data_format}."
            raise ValueError(err_msg)

    @staticmethod
    def upload_data(data_path: str) -> str:
        # Upload the data to the provider
        provider_file = openai.files.create(
            file=open(data_path, "rb"),
            purpose="fine-tune",
        )
        return provider_file.id

    @staticmethod
    def start_remote_training(
        train_file_id: str,
        model: id,
        train_kwargs: Optional[Dict[str, Any]] = None
    ) -> str:
        train_kwargs = train_kwargs or {}
        provider_job = openai.fine_tuning.jobs.create(
            model=model,
            training_file=train_file_id,
            hyperparameters=train_kwargs,
        )
        return provider_job.id

    @staticmethod
    def wait_for_job(
        job: TrainingJobOpenAI,
        poll_frequency: int = 20,
    ):
        # Poll for the job until it is done
        done = False
        cur_event_id = None
        reported_estimated_time = False
        while not done:
            # Report estimated time if not already reported
            if not reported_estimated_time:
                remote_job = openai.fine_tuning.jobs.retrieve(job.provider_job_id)
                timestamp = remote_job.estimated_finish
                if timestamp:
                    estimated_finish_dt = datetime.fromtimestamp(timestamp)
                    delta_dt = estimated_finish_dt - datetime.now()
                    print(f"[OpenAI Provider] The OpenAI estimated time remaining is: {delta_dt}")
                    reported_estimated_time = True

            # Get new events
            page = openai.fine_tuning.jobs.list_events(fine_tuning_job_id=job.provider_job_id, limit=1)
            new_event = page.data[0] if page.data else None
            if new_event and new_event.id != cur_event_id:
                dt = datetime.fromtimestamp(new_event.created_at)
                print(f"[OpenAI Provider] {dt} {new_event.message}")
                cur_event_id = new_event.id

            # Sleep and update the flag
            time.sleep(poll_frequency)
            done = OpenAIProvider.is_terminal_training_status(job.status())

    @staticmethod
    def get_trained_model(job):
        status = job.status()
        if status != TrainingStatus.succeeded:
            err_msg = f"Job status is {status}."
            err_msg += f" Must be {TrainingStatus.succeeded} to retrieve model."
            raise Exception(err_msg)

        provider_job = openai.fine_tuning.jobs.retrieve(job.provider_job_id)
        finetuned_model = provider_job.fine_tuned_model
        return finetuned_model



from typing import List

import dspy


def format_examples(examples: List[dspy.Example]) -> str:
    if isinstance(examples, str):
        return examples

    formatted_example = ""

    for example in examples:
        input_keys = example.inputs().keys()
        label_keys = example.labels().keys()

        formatted_example += "Inputs:\n"
        for key in input_keys:
            formatted_example += f"{key}: {example[key]}\n"

        formatted_example += "Outputs:\n"
        for key in label_keys:
            formatted_example += f"{key}: {example[key]}\n"

    return formatted_example


import litellm
import numpy as np


class Embedder:
    """DSPy embedding class.

    The class for computing embeddings for text inputs. This class provides a unified interface for both:

    1. Hosted embedding models (e.g. OpenAI's text-embedding-3-small) via litellm integration
    2. Custom embedding functions that you provide

    For hosted models, simply pass the model name as a string (e.g., "openai/text-embedding-3-small"). The class will use
    litellm to handle the API calls and caching.

    For custom embedding models, pass a callable function that:
    - Takes a list of strings as input.
    - Returns embeddings as either:
        - A 2D numpy array of float32 values
        - A 2D list of float32 values
    - Each row should represent one embedding vector

    Args:
        model: The embedding model to use. This can be either a string (representing the name of the hosted embedding
            model, must be an embedding model supported by litellm) or a callable that represents a custom embedding
            model.
        batch_size (int, optional): The default batch size for processing inputs in batches. Defaults to 200.
        caching (bool, optional): Whether to cache the embedding response when using a hosted model. Defaults to True.
        **kwargs: Additional default keyword arguments to pass to the embedding model.

    Examples:
        Example 1: Using a hosted model.

        ```python
        import dspy

        embedder = dspy.Embedder("openai/text-embedding-3-small", batch_size=100)
        embeddings = embedder(["hello", "world"])

        assert embeddings.shape == (2, 1536)
        ```

        Example 2: Using a custom function.

        ```python
        import dspy
        import numpy as np

        def my_embedder(texts):
            return np.random.rand(len(texts), 10)

        embedder = dspy.Embedder(my_embedder)
        embeddings = embedder(["hello", "world"], batch_size=1)

        assert embeddings.shape == (2, 10)
        ```
    """

    def __init__(self, model, batch_size=200, caching=True, **kwargs):
        self.model = model
        self.batch_size = batch_size
        self.caching = caching
        self.default_kwargs = kwargs

    def __call__(self, inputs, batch_size=None, caching=None, **kwargs):
        """Compute embeddings for the given inputs.

        Args:
            inputs: The inputs to compute embeddings for, can be a single string or a list of strings.
            batch_size (int, optional): The batch size for processing inputs. If None, defaults to the batch_size set during initialization.
            caching (bool, optional): Whether to cache the embedding response when using a hosted model. If None, defaults to the caching setting from initialization.
            **kwargs: Additional keyword arguments to pass to the embedding model. These will override the default kwargs provided during initialization.

        Returns:
            numpy.ndarray: If the input is a single string, returns a 1D numpy array representing the embedding.
            If the input is a list of strings, returns a 2D numpy array of embeddings, one embedding per row.
        """

        if isinstance(inputs, str):
            is_single_input = True
            inputs = [inputs]
        else:
            is_single_input = False

        assert all(isinstance(inp, str) for inp in inputs), "All inputs must be strings."

        if batch_size is None:
            batch_size = self.batch_size
        if caching is None:
            caching = self.caching

        merged_kwargs = self.default_kwargs.copy()
        merged_kwargs.update(kwargs)

        embeddings_list = []

        def chunk(inputs_list, size):
            for i in range(0, len(inputs_list), size):
                yield inputs_list[i : i + size]

        for batch_inputs in chunk(inputs, batch_size):
            if isinstance(self.model, str):
                embedding_response = litellm.embedding(
                    model=self.model, input=batch_inputs, caching=caching, **merged_kwargs
                )
                batch_embeddings = [data["embedding"] for data in embedding_response.data]
            elif callable(self.model):
                batch_embeddings = self.model(batch_inputs, **merged_kwargs)
            else:
                raise ValueError(
                    f"`model` in `dspy.Embedder` must be a string or a callable, but got {type(self.model)}."
                )

            embeddings_list.extend(batch_embeddings)

        embeddings = np.array(embeddings_list, dtype=np.float32)

        if is_single_input:
            return embeddings[0]
        else:
            return embeddings



import copy
import threading
from contextlib import contextmanager

from dspy.dsp.utils.utils import dotdict

DEFAULT_CONFIG = dotdict(
    lm=None,
    adapter=None,
    rm=None,
    branch_idx=0,
    trace=[],
    bypass_assert=False,
    bypass_suggest=False,
    assert_failures=0,
    suggest_failures=0,
    experimental=False,
    backoff_time=10,
    callbacks=[],
    async_max_workers=8,
    send_stream=None,
)

# Global base configuration and owner tracking
main_thread_config = copy.deepcopy(DEFAULT_CONFIG)
config_owner_thread_id = None

# Global lock for settings configuration
global_lock = threading.Lock()


class ThreadLocalOverrides(threading.local):
    def __init__(self):
        self.overrides = dotdict()


thread_local_overrides = ThreadLocalOverrides()


class Settings:
    """
    A singleton class for DSPy configuration settings.
    Thread-safe global configuration.
    - 'configure' can be called by only one 'owner' thread (the first thread that calls it).
    - Other threads see the configured global values from 'main_thread_config'.
    - 'context' sets thread-local overrides. These overrides propagate to threads spawned
      inside that context block, when (and only when!) using a ParallelExecutor that copies overrides.

      1. Only one unique thread (which can be any thread!) can call dspy.configure.
      2. It affects a global state, visible to all. As a result, user threads work, but they shouldn't be
         mixed with concurrent changes to dspy.configure from the "main" thread.
         (TODO: In the future, add warnings: if there are near-in-time user-thread reads followed by .configure calls.)
      3. Any thread can use dspy.context. It propagates to child threads created with DSPy primitives: Parallel, asyncify, etc.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @property
    def lock(self):
        return global_lock

    def __getattr__(self, name):
        overrides = getattr(thread_local_overrides, "overrides", dotdict())
        if name in overrides:
            return overrides[name]
        elif name in main_thread_config:
            return main_thread_config[name]
        else:
            raise AttributeError(f"'Settings' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name in ("_instance",):
            super().__setattr__(name, value)
        else:
            self.configure(**{name: value})

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __contains__(self, key):
        overrides = getattr(thread_local_overrides, "overrides", dotdict())
        return key in overrides or key in main_thread_config

    def get(self, key, default=None):
        try:
            return self[key]
        except AttributeError:
            return default

    def copy(self):
        overrides = getattr(thread_local_overrides, "overrides", dotdict())
        return dotdict({**main_thread_config, **overrides})

    @property
    def config(self):
        return self.copy()

    def configure(self, **kwargs):
        global main_thread_config, config_owner_thread_id
        current_thread_id = threading.get_ident()

        with self.lock:
            # First configuration: establish ownership. If ownership established, only that thread can configure.
            if config_owner_thread_id in [None, current_thread_id]:
                config_owner_thread_id = current_thread_id
            else:
                raise RuntimeError("dspy.settings can only be changed by the thread that initially configured it.")

        # Update global config
        for k, v in kwargs.items():
            main_thread_config[k] = v

    @contextmanager
    def context(self, **kwargs):
        """
        Context manager for temporary configuration changes at the thread level.
        Does not affect global configuration. Changes only apply to the current thread.
        If threads are spawned inside this block using ParallelExecutor, they will inherit these overrides.
        """

        original_overrides = getattr(thread_local_overrides, "overrides", dotdict()).copy()
        new_overrides = dotdict({**main_thread_config, **original_overrides, **kwargs})
        thread_local_overrides.overrides = new_overrides

        try:
            yield
        finally:
            thread_local_overrides.overrides = original_overrides

    def __repr__(self):
        overrides = getattr(thread_local_overrides, "overrides", dotdict())
        combined_config = {**main_thread_config, **overrides}
        return repr(combined_config)


settings = Settings()



from unittest import mock

import pydantic
import pytest
from pydantic import create_model

import dspy


def test_json_adapter_passes_structured_output_when_supported_by_model():
    class OutputField3(pydantic.BaseModel):
        subfield1: int = pydantic.Field(description="Int subfield 1", ge=0, le=10)
        subfield2: float = pydantic.Field(description="Float subfield 2")

    class TestSignature(dspy.Signature):
        input1: str = dspy.InputField()
        output1: str = dspy.OutputField()  # Description intentionally left blank
        output2: bool = dspy.OutputField(desc="Boolean output field")
        output3: OutputField3 = dspy.OutputField(desc="Nested output field")
        output4_unannotated = dspy.OutputField(desc="Unannotated output field")

    program = dspy.Predict(TestSignature)

    # Configure DSPy to use an OpenAI LM that supports structured outputs
    dspy.configure(lm=dspy.LM(model="openai/gpt4o"), adapter=dspy.JSONAdapter())
    with mock.patch("litellm.completion") as mock_completion:
        program(input1="Test input")

    def clean_schema_extra(field_name, field_info):
        attrs = dict(field_info.__repr_args__())
        if "json_schema_extra" in attrs:
            attrs["json_schema_extra"] = {
                k: v
                for k, v in attrs["json_schema_extra"].items()
                if k != "__dspy_field_type" and not (k == "desc" and v == f"${{{field_name}}}")
            }
        return attrs

    mock_completion.assert_called_once()
    _, call_kwargs = mock_completion.call_args
    response_format = call_kwargs.get("response_format")
    assert response_format is not None
    assert issubclass(response_format, pydantic.BaseModel)
    assert response_format.model_fields.keys() == {"output1", "output2", "output3", "output4_unannotated"}
    for field_name in response_format.model_fields:
        assert dict(response_format.model_fields[field_name].__repr_args__()) == clean_schema_extra(
            field_name=field_name,
            field_info=TestSignature.output_fields[field_name],
        )

    # Configure DSPy to use a model from a fake provider that doesn't support structured outputs
    dspy.configure(lm=dspy.LM(model="fakeprovider/fakemodel"), adapter=dspy.JSONAdapter())
    with mock.patch("litellm.completion") as mock_completion:
        program(input1="Test input")

    mock_completion.assert_called_once()
    _, call_kwargs = mock_completion.call_args
    assert response_format not in call_kwargs


def test_json_adapter_falls_back_when_structured_outputs_fails():
    class TestSignature(dspy.Signature):
        input1: str = dspy.InputField()
        output1: str = dspy.OutputField(desc="String output field")

    dspy.configure(lm=dspy.LM(model="openai/gpt4o"), adapter=dspy.JSONAdapter())
    program = dspy.Predict(TestSignature)
    with mock.patch("litellm.completion") as mock_completion:
        mock_completion.side_effect = [Exception("Bad structured outputs!"), mock_completion.return_value]
        program(input1="Test input")
        assert mock_completion.call_count == 2
        _, first_call_kwargs = mock_completion.call_args_list[0]
        assert issubclass(first_call_kwargs.get("response_format"), pydantic.BaseModel)
        _, second_call_kwargs = mock_completion.call_args_list[1]
        assert second_call_kwargs.get("response_format") == {"type": "json_object"}


def test_json_adapter_with_structured_outputs_does_not_mutate_original_signature():
    class OutputField3(pydantic.BaseModel):
        subfield1: int = pydantic.Field(description="Int subfield 1")
        subfield2: float = pydantic.Field(description="Float subfield 2")

    class TestSignature(dspy.Signature):
        input1: str = dspy.InputField()
        output1: str = dspy.OutputField()  # Description intentionally left blank
        output2: bool = dspy.OutputField(desc="Boolean output field")
        output3: OutputField3 = dspy.OutputField(desc="Nested output field")
        output4_unannotated = dspy.OutputField(desc="Unannotated output field")

    dspy.configure(lm=dspy.LM(model="openai/gpt4o"), adapter=dspy.JSONAdapter())
    program = dspy.Predict(TestSignature)
    with mock.patch("litellm.completion"):
        program(input1="Test input")

    assert program.signature.output_fields == TestSignature.output_fields



import dspy

from copy import deepcopy
from pydantic.fields import FieldInfo

from dspy.predict.avatar.signatures import Actor
from dspy.signatures.signature import ensure_signature
from dspy.predict.avatar.models import Action, ActionOutput, Tool


def get_number_with_suffix(number: int) -> str:
    if number == 1:
        return "1st"
    elif number == 2:
        return "2nd"
    elif number == 3:
        return "3rd"
    else:
        return f"{number}th"


class Avatar(dspy.Module):
    def __init__(
        self,
        signature,
        tools,
        max_iters=3,
        verbose=False,
    ):
        self.signature = ensure_signature(signature)
        self.input_fields = self.signature.input_fields
        self.output_fields = self.signature.output_fields

        self.finish_tool = Tool(
            tool=None,
            name="Finish",
            desc="returns the final output and finishes the task",
        )

        self.tools = tools + [self.finish_tool]
        self.actor_signature = Actor

        for field in list(self.input_fields.keys())[::-1]:
            self.actor_signature = self.actor_signature.append(
                field,
                self._get_field(self.input_fields[field]),
                type_=self.input_fields[field].annotation,
            )

        self.verbose = verbose
        self.max_iters = max_iters
        self.actor = dspy.TypedPredictor(self.actor_signature)

        self.actor_clone = deepcopy(self.actor)


    def _get_field(self, field_info: FieldInfo):
        if field_info.json_schema_extra['__dspy_field_type'] == 'input':
            return dspy.InputField(
                prefix=field_info.json_schema_extra['prefix'],
                desc=field_info.json_schema_extra['desc'],
                format=field_info.json_schema_extra['format'] if 'format' in field_info.json_schema_extra else None,
            )
        elif field_info.json_schema_extra['__dspy_field_type'] == 'output':
            return dspy.OutputField(
                prefix=field_info.json_schema_extra['prefix'],
                desc=field_info.json_schema_extra['desc'],
                format=field_info.json_schema_extra['format'] if 'format' in field_info.json_schema_extra else None,
            )
        else:
            raise ValueError(f"Unknown field type: {field_info.json_schema_extra['__dspy_field_type']}")


    def _update_signature(self, idx: int, omit_action: bool = False):
        self.actor.signature = self.actor.signature.with_updated_fields(
            f"action_{idx}", 
            Action, 
            __dspy_field_type="input"
        )

        self.actor.signature = self.actor.signature.append(
            f"result_{idx}",
            dspy.InputField(
                prefix=f"Result {idx}:",
                desc=f"{get_number_with_suffix(idx)} result",
                type_=str,
            )
        )
        
        if omit_action:
            for field in list(self.output_fields.keys()):
                self.actor.signature = self.actor.signature.append(
                    field,
                    self._get_field(self.output_fields[field]),
                    type_=self.output_fields[field].annotation,
                )
        else:        
            self.actor.signature = self.actor.signature.append(
                f"action_{idx+1}",
                dspy.OutputField(
                    prefix=f"Action {idx+1}:",
                    desc=f"{get_number_with_suffix(idx+1)} action to taken",
                )
            )
            self.actor.signature = self.actor.signature.with_updated_fields(
                f"action_{idx+1}",
                Action,
            )


    def _call_tool(self, tool_name: str, tool_input_query: str) -> str:
        for tool in self.tools:
            if tool.name == tool_name:
                return tool.tool.run(tool_input_query)


    def forward(self, **kwargs):
        if self.verbose:
            print("Starting the task...")
        
        args = {
            "goal" : self.signature.__doc__,
            "tools" : [tool.name for tool in self.tools],
        }
        
        for key in self.input_fields.keys():
            if key in kwargs:
                args[key] = kwargs[key]
        
        idx = 1
        tool_name = None
        action_results: list[ActionOutput] = []
        max_iters = None if "max_iters" not in kwargs else kwargs["max_iters"]

        while tool_name != "Finish" and (max_iters > 0 if max_iters else True):
            actor_output = self.actor(**args)
            action = getattr(actor_output, f"action_{idx}")

            tool_name = action.tool_name
            tool_input_query = action.tool_input_query

            if self.verbose:
                print(f"Action {idx}: {tool_name} ({tool_input_query})")

            if tool_name != "Finish":
                tool_output = self._call_tool(tool_name, tool_input_query)
                action_results.append(
                    ActionOutput(
                        tool_name=tool_name, 
                        tool_input_query=tool_input_query, 
                        tool_output=tool_output
                    )
                )

                self._update_signature(idx)

                args[f"action_{idx}"] = action
                args[f"result_{idx}"] = tool_output
            else:
                self._update_signature(idx, omit_action=True)

                args[f"action_{idx}"] = action
                args[f"result_{idx}"] = "Gathered all information needed to finish the task."
                break

            idx += 1

            if max_iters:
                max_iters -= 1

        final_answer = self.actor(**args)
        self.actor = deepcopy(self.actor_clone)

        return dspy.Prediction(
            **{key: getattr(final_answer, key) for key in self.output_fields.keys()},
            actions=action_results,
        )



#replace_package_name_marker
name="test-pkg-hmoazam"
#replace_package_version_marker
__version__="5.11.14"
__description__="DSPy"
__url__="https://github.com/stanfordnlp/dspy"
__author__="Omar Khattab"
__author_email__="okhattab@stanford.edu"


import random

from pydantic import BaseModel

from dspy.adapters.image_utils import Image
from dspy.predict.parameter import Parameter
from dspy.primitives.prediction import Prediction
from dspy.primitives.program import Module
from dspy.signatures.signature import ensure_signature
from dspy.utils.callback import with_callbacks


class Predict(Module, Parameter):
    def __init__(self, signature, callbacks=None, **config):
        self.stage = random.randbytes(8).hex()
        self.signature = ensure_signature(signature)
        self.config = config
        self.callbacks = callbacks or []
        self.reset()

    def reset(self):
        self.lm = None
        self.traces = []
        self.train = []
        self.demos = []

    def dump_state(self):
        state_keys = ["lm", "traces", "train"]
        state = {k: getattr(self, k) for k in state_keys}

        state["demos"] = []
        for demo in self.demos:
            demo = demo.copy()

            for field in demo:
                # FIXME: Saving BaseModels as strings in examples doesn't matter because you never re-access as an object
                # It does matter for images
                if isinstance(demo[field], Image):
                    demo[field] = demo[field].model_dump()
                elif isinstance(demo[field], BaseModel):
                    demo[field] = demo[field].model_dump_json()

            state["demos"].append(demo)

        state["signature"] = self.signature.dump_state()
        return state

    def load_state(self, state):
        """Load the saved state of a `Predict` object.

        Args:
            state (dict): The saved state of a `Predict` object.

        Returns:
            self: Returns self to allow method chaining
        """
        excluded_keys = ["signature", "extended_signature"]
        for name, value in state.items():
            # `excluded_keys` are fields that go through special handling.
            if name not in excluded_keys:
                setattr(self, name, value)

        # FIXME: Images are getting special treatment, but all basemodels initialized from json should be converted back to objects
        for demo in self.demos:
            for field in demo:
                if isinstance(demo[field], dict) and "url" in demo[field]:
                    url = demo[field]["url"]
                    if not isinstance(url, str):
                        raise ValueError(f"Image URL must be a string, got {type(url)}")
                    demo[field] = Image(url=url)

        self.signature = self.signature.load_state(state["signature"])

        if "extended_signature" in state: # legacy, up to and including 2.5, for CoT.
            raise NotImplementedError("Loading extended_signature is no longer supported in DSPy 2.6+")

        return self

    @with_callbacks
    def __call__(self, **kwargs):
        return self.forward(**kwargs)

    def forward(self, **kwargs):
        import dspy

        # Extract the three privileged keyword arguments.
        assert "new_signature" not in kwargs, "new_signature is no longer a valid keyword argument."
        signature = ensure_signature(kwargs.pop("signature", self.signature))
        demos = kwargs.pop("demos", self.demos)
        config = dict(**self.config, **kwargs.pop("config", {}))

        # Get the right LM to use.
        lm = kwargs.pop("lm", self.lm) or dspy.settings.lm
        assert isinstance(lm, dspy.LM), "No LM is loaded."

        # If temperature is 0.0 but its n > 1, set temperature to 0.7.
        temperature = config.get("temperature")
        temperature = lm.kwargs["temperature"] if temperature is None else temperature
        num_generations = config.get("n") or lm.kwargs.get("n") or lm.kwargs.get("num_generations") or 1

        if (temperature is None or temperature <= 0.15) and num_generations > 1:
            config["temperature"] = 0.7

        if not all(k in kwargs for k in signature.input_fields):
            present = [k for k in signature.input_fields if k in kwargs]
            missing = [k for k in signature.input_fields if k not in kwargs]
            print(f"WARNING: Not all input fields were provided to module. Present: {present}. Missing: {missing}.")

        import dspy
        adapter = dspy.settings.adapter or dspy.ChatAdapter()
        completions = adapter(lm, lm_kwargs=config, signature=signature, demos=demos, inputs=kwargs)

        pred = Prediction.from_completions(completions, signature=signature)

        if kwargs.pop("_trace", True) and dspy.settings.trace is not None:
            trace = dspy.settings.trace
            trace.append((self, {**kwargs}, pred))

        return pred

    def update_config(self, **kwargs):
        self.config = {**self.config, **kwargs}

    def get_config(self):
        return self.config

    def __repr__(self):
        return f"{self.__class__.__name__}({self.signature})"


# TODO: get some defaults during init from the context window?
# # TODO: FIXME: Hmm, I guess expected behavior is that contexts can
# affect execution. Well, we need to determine whether context dominates, __init__ demoninates, or forward dominates.
# Generally, unless overwritten, we'd see n=None, temperature=None.
# That will eventually mean we have to learn them.



from dspy.predict.aggregation import majority
from dspy.predict.chain_of_thought import ChainOfThought
from dspy.predict.chain_of_thought_with_hint import ChainOfThoughtWithHint
from dspy.predict.knn import KNN
from dspy.predict.multi_chain_comparison import MultiChainComparison
from dspy.predict.predict import Predict
from dspy.predict.program_of_thought import ProgramOfThought
from dspy.predict.react import ReAct, Tool
from dspy.predict.parallel import Parallel

__all__ = [
    "majority",
    "ChainOfThought",
    "ChainOfThoughtWithHint",
    "KNN",
    "MultiChainComparison",
    "Predict",
    "ProgramOfThought",
    "ReAct",
    "Tool",
    "Parallel",
]



import inspect
import logging
import math
import os
import random
import shutil
import sys

import numpy as np

try:
    from IPython.core.magics.code import extract_symbols
except ImportError:
    # Won't be able to read code from juptyer notebooks
    extract_symbols = None

import dspy
from dspy.teleprompt.bootstrap import BootstrapFewShot, LabeledFewShot

"""
This file consists of helper functions for our variety of optimizers.
"""

### OPTIMIZER TRAINING UTILS ###


def create_minibatch(trainset, batch_size=50, rng=None):
    """Create a minibatch from the trainset."""

    # Ensure batch_size isn't larger than the size of the dataset
    batch_size = min(batch_size, len(trainset))

    # If no RNG is provided, fall back to the global random instance
    rng = rng or random

    # Randomly sample indices for the mini-batch using the provided rng
    sampled_indices = rng.sample(range(len(trainset)), batch_size)

    # Create the mini-batch using the sampled indices
    minibatch = [trainset[i] for i in sampled_indices]

    return minibatch


def eval_candidate_program(batch_size, trainset, candidate_program, evaluate, rng=None):
    """Evaluate a candidate program on the trainset, using the specified batch size."""

    try:
        # Evaluate on the full trainset
        if batch_size >= len(trainset):
            score = evaluate(candidate_program, devset=trainset)
        # Or evaluate on a minibatch
        else:
            score = evaluate(
                candidate_program,
                devset=create_minibatch(trainset, batch_size, rng),
            )
    except Exception as e:
        print(f"Exception occurred: {e}")
        score = 0.0  # TODO: Handle this better, as -ve scores are possible

    return score


def eval_candidate_program_with_pruning(
    trial, trial_logs, trainset, candidate_program, evaluate, trial_num, batch_size=100,
):
    """Evaluation of candidate_program with pruning implemented"""

    # Evaluate with the new prompts
    total_score = 0
    num_batches = math.ceil(len(trainset) / batch_size)
    total_eval_size = 0

    for i in range(num_batches):
        start_index = i * batch_size
        end_index = min((i + 1) * batch_size, len(trainset))
        split_trainset = trainset[start_index:end_index]
        split_score = evaluate(
            candidate_program, devset=split_trainset, display_table=0,
        )
        print(f"{i}st split score: {split_score}")
        total_eval_size += len(split_trainset)

        total_score += split_score * len(split_trainset)
        curr_weighted_avg_score = total_score / min((i + 1) * batch_size, len(trainset))
        print(f"curr average score: {curr_weighted_avg_score}")

        trial.report(curr_weighted_avg_score, i)

        # Handle pruning based on the intermediate value.
        if trial.should_prune():
            print("Trial pruned.")
            trial_logs[trial_num]["score"] = curr_weighted_avg_score
            trial_logs[trial_num]["num_eval_calls"] = total_eval_size
            trial_logs[trial_num]["pruned"] = True
            return curr_weighted_avg_score, trial_logs, total_eval_size, True

    print(f"Fully evaled score: {curr_weighted_avg_score}")
    score = curr_weighted_avg_score

    trial_logs[trial_num]["full_eval"] = False
    trial_logs[trial_num]["score"] = score
    trial_logs[trial_num]["pruned"] = False
    return score, trial_logs, total_eval_size, False


def get_program_with_highest_avg_score(param_score_dict, fully_evaled_param_combos):
    """Used as a helper function for bayesian + minibatching optimizers. Returns the program with the highest average score from the batches evaluated so far."""

    # Calculate the mean for each combination of categorical parameters, based on past trials
    results = []
    for key, values in param_score_dict.items():
        scores = np.array([v[0] for v in values])
        mean = np.average(scores)
        program = values[0][1]
        results.append((key, mean, program))

    # Sort results by the mean
    sorted_results = sorted(results, key=lambda x: x[1], reverse=True)

    # Find the combination with the highest mean, skip fully evaluated ones
    for combination in sorted_results:
        key, mean, program = combination

        if key in fully_evaled_param_combos:
            continue

        return program, mean, key

    # If no valid program is found, we return the last valid one that we found
    return program, mean, key


def calculate_last_n_proposed_quality(
    base_program, trial_logs, evaluate, trainset, devset, n,
):
    """
    Calculate the average and best quality of the last n programs proposed. This is useful for seeing if our proposals
    are actually 'improving' overtime or not.
    """
    # Get the trials from the last n keys in trial logs
    last_n_trial_nums = list(trial_logs.keys())[-n:]

    # Calculate the average and best score of these trials
    # if num_eval_calls in the trial is less than the trainset, throw a not-implemented error for now
    total_train_score = 0
    best_train_score = 0
    total_dev_score = 0
    best_dev_score = 0
    for trial_num in last_n_trial_nums:
        full_eval = trial_logs[trial_num]["full_eval"]
        if not full_eval:
            raise NotImplementedError(
                "Still need to implement non full eval handling in calculate_last_n_proposed_quality",
            )
        train_score = trial_logs[trial_num]["score"]
        program = base_program.deepcopy()
        program.load(trial_logs[trial_num]["program_path"])

        dev_score = evaluate(program, devset=devset)

        total_train_score += train_score
        total_dev_score += dev_score
        if train_score > best_train_score:
            best_train_score = train_score
            best_dev_score = dev_score

    return best_train_score, total_train_score / n, best_dev_score, total_dev_score / n


### LOGGING UTILS ###


def get_task_model_history_for_full_example(
    candidate_program, task_model, devset, evaluate,
):
    """Get a full trace of the task model's history for a given candidate program."""
    _ = evaluate(candidate_program, devset=devset[:1])
    _ = task_model.inspect_history(n=len(candidate_program.predictors()))
    return task_model.inspect_history(n=len(candidate_program.predictors()))


def print_full_program(program):
    """Print out the program's instructions & prefixes for each module."""
    for i, predictor in enumerate(program.predictors()):
        print(f"Predictor {i}")
        print(f"i: {get_signature(predictor).instructions}")
        *_, last_field = get_signature(predictor).fields.values()
        print(f"p: {last_field.json_schema_extra['prefix']}")
    print("\n")


def save_candidate_program(program, log_dir, trial_num, note=None):
    """Save the candidate program to the log directory."""

    if log_dir is None:
        return None

    # Ensure the directory exists
    eval_programs_dir = os.path.join(log_dir, "evaluated_programs")
    os.makedirs(eval_programs_dir, exist_ok=True)

    # Define the save path for the program
    if note:
        save_path = os.path.join(eval_programs_dir, f"program_{trial_num}_{note}")
    else:
        save_path = os.path.join(eval_programs_dir, f"program_{trial_num}")

    # Save the program
    program.save(save_path)

    return save_path


def save_file_to_log_dir(source_file_path, log_dir):
    if log_dir is None:
        return
    """Save a file to our log directory"""
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    destination_file_path = os.path.join(log_dir, os.path.basename(source_file_path))

    # Copy the file
    shutil.copy(source_file_path, destination_file_path)


def setup_logging(log_dir):
    """Setup logger, which will log our print statements to a txt file at our log_dir for later viewing"""
    if log_dir is None:
        return
    # Create a logger
    logger = logging.getLogger()
    logger.setLevel(logging.WARNING)

    # Create a file handler that logs debug and higher level messages
    file_handler = logging.FileHandler(f"{log_dir}/logs.txt")
    file_handler.setLevel(logging.WARNING)
    file_formatter = logging.Formatter("%(asctime)s - %(message)s")
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    # Create a console handler with a higher log level
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.WARNING)
    console_formatter = logging.Formatter("%(message)s")
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)


### OTHER UTILS ###

def get_prompt_model(prompt_model):
    if prompt_model:
        return prompt_model
    else:
        return dspy.settings.lm

def get_signature(predictor):
    assert hasattr(predictor, "signature")
    return predictor.signature


def set_signature(predictor, updated_signature):
    assert hasattr(predictor, "signature")
    predictor.signature = updated_signature


def create_n_fewshot_demo_sets(
    student,
    num_candidate_sets,
    trainset,
    max_labeled_demos,
    max_bootstrapped_demos,
    metric,
    teacher_settings,
    max_errors=10,
    max_rounds=1,
    labeled_sample=True,
    min_num_samples=1,
    metric_threshold=None,
    teacher=None,
    include_non_bootstrapped=True,
    seed=0,
    rng=None
):
    """
    This function is copied from random_search.py, and creates fewshot examples in the same way that random search does.
    This allows us to take advantage of using the same fewshot examples when we use the same random seed in our optimizers.
    """
    demo_candidates = {}

    # Account for confusing way this is set up, where we add in 3 more candidate sets to the N specified
    num_candidate_sets -= 3

    # Initialize demo_candidates dictionary
    for i, _ in enumerate(student.predictors()):
        demo_candidates[i] = []

    rng = rng or random.Random(seed)

    # Go through and create each candidate set
    for seed in range(-3, num_candidate_sets):

        print(f"Bootstrapping set {seed+4}/{num_candidate_sets+3}")

        trainset_copy = list(trainset)

        if seed == -3 and include_non_bootstrapped:
            # zero-shot
            program2 = student.reset_copy()

        elif (
            seed == -2
            and max_labeled_demos > 0
            and include_non_bootstrapped
        ):
            # labels only
            teleprompter = LabeledFewShot(k=max_labeled_demos)
            program2 = teleprompter.compile(
                student, trainset=trainset_copy, sample=labeled_sample,
            )

        elif seed == -1:
            # unshuffled few-shot
            program = BootstrapFewShot(
                metric=metric,
                max_errors=max_errors,
                max_bootstrapped_demos=max_bootstrapped_demos,
                max_labeled_demos=max_labeled_demos,
                teacher_settings=teacher_settings,
                max_rounds=max_rounds,
            )
            program2 = program.compile(student, teacher=teacher, trainset=trainset_copy)

        else:
            # shuffled few-shot
            rng.shuffle(trainset_copy)
            size = rng.randint(min_num_samples, max_bootstrapped_demos)

            teleprompter = BootstrapFewShot(
                metric=metric,
                max_errors=max_errors,
                metric_threshold=metric_threshold,
                max_bootstrapped_demos=size,
                max_labeled_demos=max_labeled_demos,
                teacher_settings=teacher_settings,
                max_rounds=max_rounds,
            )

            program2 = teleprompter.compile(
                student, teacher=teacher, trainset=trainset_copy,
            )

        for i, _ in enumerate(student.predictors()):
            demo_candidates[i].append(program2.predictors()[i].demos)

    return demo_candidates

def old_getfile(object):
    """Work out which source or compiled file an object was defined in."""
    if inspect.ismodule(object):
        if getattr(object, '__file__', None):
            return object.__file__
        raise TypeError('{!r} is a built-in module'.format(object))
    if inspect.isclass(object):
        if hasattr(object, '__module__'):
            module = sys.modules.get(object.__module__)
            if getattr(module, '__file__', None):
                return module.__file__
            if object.__module__ == '__main__':
                raise OSError('source code not available')
        raise TypeError('{!r} is a built-in class'.format(object))
    if inspect.ismethod(object):
        object = object.__func__
    if inspect.isfunction(object):
        object = object.__code__
    if inspect.istraceback(object):
        object = object.tb_frame
    if inspect.isframe(object):
        object = object.f_code
    if inspect.iscode(object):
        return object.co_filename
    raise TypeError('module, class, method, function, traceback, frame, or '
                    'code object was expected, got {}'.format(
                    type(object).__name__))

def new_getfile(object):
    if not inspect.isclass(object):
        return old_getfile(object)
    
    # Lookup by parent module (as in current inspect)
    if hasattr(object, '__module__'):
        object_ = sys.modules.get(object.__module__)
        if hasattr(object_, '__file__'):
            return object_.__file__
    
    # If parent module is __main__, lookup by methods (NEW)
    for name, member in inspect.getmembers(object):
        if inspect.isfunction(member) and object.__qualname__ + '.' + member.__name__ == member.__qualname__:
            return inspect.getfile(member)
    raise TypeError(f'Source for {object!r} not found')

inspect.getfile = new_getfile


import functools
from typing import Any, List, Optional, Union

import requests

from dspy.dsp.cache_utils import CacheMemory, NotebookCacheMemory
from dspy.dsp.utils import dotdict

# TODO: Ideally, this takes the name of the index and looks up its port.


class ColBERTv2:
    """Wrapper for the ColBERTv2 Retrieval."""

    def __init__(
        self,
        url: str = "http://0.0.0.0",
        port: Optional[Union[str, int]] = None,
        post_requests: bool = False,
    ):
        self.post_requests = post_requests
        self.url = f"{url}:{port}" if port else url

    def __call__(
        self, query: str, k: int = 10, simplify: bool = False,
    ) -> Union[list[str], list[dotdict]]:
        if self.post_requests:
            topk: list[dict[str, Any]] = colbertv2_post_request(self.url, query, k)
        else:
            topk: list[dict[str, Any]] = colbertv2_get_request(self.url, query, k)

        if simplify:
            return [psg["long_text"] for psg in topk]

        return [dotdict(psg) for psg in topk]


@CacheMemory.cache
def colbertv2_get_request_v2(url: str, query: str, k: int):
    assert (
        k <= 100
    ), "Only k <= 100 is supported for the hosted ColBERTv2 server at the moment."

    payload = {"query": query, "k": k}
    res = requests.get(url, params=payload, timeout=10)

    topk = res.json()["topk"][:k]
    topk = [{**d, "long_text": d["text"]} for d in topk]
    return topk[:k]


@functools.cache
@NotebookCacheMemory.cache
def colbertv2_get_request_v2_wrapped(*args, **kwargs):
    return colbertv2_get_request_v2(*args, **kwargs)


colbertv2_get_request = colbertv2_get_request_v2_wrapped


@CacheMemory.cache
def colbertv2_post_request_v2(url: str, query: str, k: int):
    headers = {"Content-Type": "application/json; charset=utf-8"}
    payload = {"query": query, "k": k}
    res = requests.post(url, json=payload, headers=headers, timeout=10)

    return res.json()["topk"][:k]


@functools.cache
@NotebookCacheMemory.cache
def colbertv2_post_request_v2_wrapped(*args, **kwargs):
    return colbertv2_post_request_v2(*args, **kwargs)


colbertv2_post_request = colbertv2_post_request_v2_wrapped

class ColBERTv2RetrieverLocal:
    def __init__(self,passages:List[str],colbert_config=None,load_only:bool=False):
        """Colbertv2 retriever module

        Args:
            passages (List[str]): list of passages
            colbert_config (ColBERTConfig, optional): colbert config for building and searching. Defaults to None.
            load_only (bool, optional): whether to load the index or build and then load. Defaults to False.
        """
        assert colbert_config is not None, "Please pass a valid colbert_config, which you can import from colbert.infra.config import ColBERTConfig and modify it"
        self.colbert_config = colbert_config

        assert self.colbert_config.checkpoint is not None, "Please pass a valid checkpoint like colbert-ir/colbertv2.0, which you can modify in the ColBERTConfig with attribute name checkpoint"
        self.passages = passages
        
        assert self.colbert_config.index_name is not None, "Please pass a valid index_name, which you can modify in the ColBERTConfig with attribute name index_name"
        self.passages = passages

        if not load_only:
            print(f"Building the index for experiment {self.colbert_config.experiment} with index name {self.colbert_config.index_name}")
            self.build_index()
        
        print(f"Loading the index for experiment {self.colbert_config.experiment} with index name {self.colbert_config.index_name}")
        self.searcher = self.get_index()

    def build_index(self):

        try:
            import colbert
        except ImportError:
            print("Colbert not found. Please check your installation or install the module using pip install colbert-ai[faiss-gpu,torch].")

        from colbert import Indexer
        from colbert.infra import Run, RunConfig
        with Run().context(RunConfig(nranks=self.colbert_config.nranks, experiment=self.colbert_config.experiment)):  
            indexer = Indexer(checkpoint=self.colbert_config.checkpoint, config=self.colbert_config)
            indexer.index(name=self.colbert_config.index_name, collection=self.passages, overwrite=True)

    def get_index(self):
        try:
            import colbert
        except ImportError:
            print("Colbert not found. Please check your installation or install the module using pip install colbert-ai[faiss-gpu,torch].")

        from colbert import Searcher
        from colbert.infra import Run, RunConfig
        
        with Run().context(RunConfig(experiment=self.colbert_config.experiment)):
            searcher = Searcher(index=self.colbert_config.index_name, collection=self.passages)
        return searcher
    
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.forward(*args, **kwargs)

    def forward(self,query:str,k:int=7,**kwargs):
        import torch
        
        if kwargs.get("filtered_pids"):
            filtered_pids = kwargs.get("filtered_pids")
            assert type(filtered_pids) == List[int], "The filtered pids should be a list of integers"
            device = "cuda" if torch.cuda.is_available() else "cpu"
            results = self.searcher.search(
                query,
                #Number of passages to receive
                k=k, 
                #Passing the filter function of relevant 
                filter_fn=lambda pids: torch.tensor(
                    [pid for pid in pids if pid in filtered_pids],dtype=torch.int32).to(device))
        else:
            searcher_results = self.searcher.search(query, k=k)
        results = []
        for pid,rank,score in zip(*searcher_results):
            results.append(dotdict({'long_text':self.searcher.collection[pid],'score':score,'pid':pid}))
        return results

class ColBERTv2RerankerLocal:
    
    def __init__(self,colbert_config=None,checkpoint:str='bert-base-uncased'):
        try:
            import colbert
        except ImportError:
            print("Colbert not found. Please check your installation or install the module using pip install colbert-ai[faiss-gpu,torch].")
        """_summary_

        Args:
            colbert_config (ColBERTConfig, optional): Colbert config. Defaults to None.
            checkpoint_name (str, optional): checkpoint for embeddings. Defaults to 'bert-base-uncased'.
        """
        self.colbert_config = colbert_config
        self.checkpoint = checkpoint
        self.colbert_config.checkpoint = checkpoint

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.forward(*args, **kwargs)

    def forward(self,query:str,passages:List[str]=[]):
        assert len(passages) > 0, "Passages should not be empty"

        import numpy as np
        from colbert.modeling.colbert import ColBERT
        from colbert.modeling.tokenization.doc_tokenization import DocTokenizer
        from colbert.modeling.tokenization.query_tokenization import QueryTokenizer
        
        self.colbert_config.nway = len(passages)
        query_tokenizer = QueryTokenizer(self.colbert_config,verbose=1)
        doc_tokenizer = DocTokenizer(self.colbert_config)
        query_ids,query_masks = query_tokenizer.tensorize([query])
        doc_ids,doc_masks = doc_tokenizer.tensorize(passages)

        col = ColBERT(self.checkpoint,self.colbert_config) 
        Q = col.query(query_ids,query_masks)
        DOC_IDS,DOC_MASKS = col.doc(doc_ids,doc_masks,keep_dims='return_mask')
        Q_duplicated = Q.repeat_interleave(len(passages), dim=0).contiguous()
        tensor_scores = col.score(Q_duplicated,DOC_IDS,DOC_MASKS)
        passage_score_arr = np.array([score.cpu().detach().numpy().tolist() for score in tensor_scores])
        return passage_score_arr


import random
import re


class MATH:
    def __init__(self, subset):
        from datasets import load_dataset

        import dspy

        ds = load_dataset("lighteval/MATH", subset)

        # NOTE: Defaults to sub-splitting MATH's 'test' split into train/dev/test, presuming that current
        # LMs are trained on MATH's train. Makes no difference for gpt-4o-mini, but might for other models.

        dataset = [
            dspy.Example(
                question=example["problem"], reasoning=example["solution"], answer=extract_answer(example["solution"])
            ).with_inputs("question")
            for example in ds["test"]
        ]

        size = min(350, len(dataset) // 3)
        random.Random(0).shuffle(dataset)
        self.train, self.dev, self.test = dataset[:size], dataset[size : 2 * size], dataset[2 * size :]

    def metric(self, example, pred, trace=None):
        try:
            import math_equivalence
        except ImportError:
            raise ImportError("MATH's metric requires `pip install git+https://github.com/hendrycks/math.git`")

        return math_equivalence.is_equiv(example.answer, pred.answer)


def extract_answer(s):
    start = s.find("\\boxed{")
    if start == -1:
        return None

    idx = start + len("\\boxed{")
    brace_level = 1

    answer = ""
    while idx < len(s) and brace_level > 0:
        c = s[idx]
        if c == "{":
            brace_level += 1
        elif c == "}":
            brace_level -= 1
            if brace_level == 0:
                break
        answer += c
        idx += 1

    answer = re.sub(r"\\text\{[^}]*\}", "", answer)
    answer = re.sub(r"\\!", "", answer)
    return answer.strip()


"""
NOTE: MATH's official math_equivalence.is_equiv does not seem to have perfect recall.
Consider its behavior on reference values like `left[\frac{1}{2}, \frac{4}{3}\right]`.
"""



import numpy as np


class KNN:
    def __init__(self, k: int, trainset: list, vectorizer=None):
        """
        A k-nearest neighbors retriever that finds similar examples from a training set.

        Args:
            k: Number of nearest neighbors to retrieve
            trainset: List of training examples to search through
            vectorizer: Optional dspy.Embedder for computing embeddings. If None, uses sentence-transformers.

        Example:
            >>> trainset = [dsp.Example(input="hello", output="world"), ...]
            >>> knn = KNN(k=3, trainset=trainset)
            >>> similar_examples = knn(input="hello")
        """
        
        import dspy.dsp as dsp
        import dspy

        self.k = k
        self.trainset = trainset
        self.embedding = vectorizer or dspy.Embedder(dsp.SentenceTransformersVectorizer())
        trainset_casted_to_vectorize = [
            " | ".join([f"{key}: {value}" for key, value in example.items() if key in example._input_keys])
            for example in self.trainset
        ]
        self.trainset_vectors = self.embedding(trainset_casted_to_vectorize).astype(np.float32)

    def __call__(self, **kwargs) -> list:
        input_example_vector = self.embedding([" | ".join([f"{key}: {val}" for key, val in kwargs.items()])])
        scores = np.dot(self.trainset_vectors, input_example_vector.T).squeeze()
        nearest_samples_idxs = scores.argsort()[-self.k :][::-1]
        train_sampled = [self.trainset[cur_idx] for cur_idx in nearest_samples_idxs]
        return train_sampled



# import re
# from copy import deepcopy
# from typing import Any, Callable, Dict, List, Optional

# from llama_index.core.base.llms.base import BaseLLM
# from llama_index.core.base.llms.generic_utils import (
#     prompt_to_messages,
# )
# from llama_index.core.base.llms.types import ChatMessage
# from llama_index.core.base.query_pipeline.query import InputKeys, OutputKeys, QueryComponent
# from llama_index.core.callbacks.base import CallbackManager
# from llama_index.core.prompts import BasePromptTemplate, PromptTemplate
# from llama_index.core.query_pipeline import QueryPipeline

# import dspy.dsp as dsp
# import dspy
# from dspy import Predict
# from dspy.signatures.field import InputField, OutputField
# from dspy.signatures.signature import ensure_signature, make_signature, signature_to_template


# def get_formatted_template(predict_module: Predict, kwargs: Dict[str, Any]) -> str:
#     """Get formatted template from predict module."""
#     # Extract the three privileged keyword arguments.
#     signature = ensure_signature(predict_module.signature)
#     demos = predict_module.demos

#     # All of the other kwargs are presumed to fit a prefix of the signature.
#     # That is, they are input variables for the bottom most generation, so
#     # we place them inside the input - x - together with the demos.
#     x = dsp.Example(demos=demos, **kwargs)

#     # Switch to legacy format for dsp.generate
#     template = signature_to_template(signature)

#     return template(x)


# def replace_placeholder(text: str) -> str:
#     # Use a regular expression to find and replace ${...} with ${{...}}
#     return re.sub(r'\$\{([^\{\}]*)\}', r'${{\1}}', text)


# def _input_keys_from_template(template: dsp.Template) -> InputKeys:
#     """Get input keys from template."""
#     # get only fields that are marked OldInputField and NOT OldOutputField
#     # template_vars = list(template.kwargs.keys())
#     return [
#         k for k, v in template.kwargs.items() if isinstance(v, dspy.signatures.OldInputField)
#     ]

# def _output_keys_from_template(template: dsp.Template) -> InputKeys:
#     """Get output keys from template."""
#     # get only fields that are marked OldOutputField and NOT OldInputField
#     # template_vars = list(template.kwargs.keys())
#     return [
#         k for k, v in template.kwargs.items() if isinstance(v, dspy.signatures.OldOutputField)
#     ]


# class DSPyPromptTemplate(BasePromptTemplate):
#     """A prompt template for DSPy.

#     Takes in a predict module from DSPy (whether unoptimized or optimized),
#     and extracts the relevant prompt template from it given the input.
    
#     """

#     predict_module: Predict

#     def __init__(
#         self,
#         predict_module: Predict,
#         metadata: Optional[Dict[str, Any]] = None,
#         template_var_mappings: Optional[Dict[str, Any]] = None,
#         function_mappings: Optional[Dict[str, Callable]] = None,
#         **kwargs: Any,
#     ) -> None:
#         template = signature_to_template(predict_module.signature)
#         template_vars = _input_keys_from_template(template)
#         # print(f"TEMPLATE VARS: {template_vars}")
#         # raise Exception

#         super().__init__(
#             predict_module=predict_module,
#             metadata=metadata or {},
#             template_vars=template_vars,
#             kwargs=kwargs,
#             template_var_mappings=template_var_mappings,
#             function_mappings=function_mappings,
#         )
    
#     def partial_format(self, **kwargs: Any) -> "BasePromptTemplate":
#         """Returns a new prompt template with the provided kwargs."""
#         # NOTE: this is a copy of the implementation in `PromptTemplate`
#         output_parser = self.output_parser
#         self.output_parser = None

#         # get function and fixed kwargs, and add that to a copy
#         # of the current prompt object
#         prompt = deepcopy(self)
#         prompt.kwargs.update(kwargs)

#         # NOTE: put the output parser back
#         prompt.output_parser = output_parser
#         self.output_parser = output_parser
#         return prompt

#     def format(self, llm: Optional[BaseLLM] = None, **kwargs: Any) -> str:
#         """Formats the prompt template."""
#         mapped_kwargs = self._map_all_vars(kwargs)
#         return get_formatted_template(self.predict_module, mapped_kwargs)

#     def format_messages(
#         self, llm: Optional[BaseLLM] = None, **kwargs: Any,
#     ) -> List[ChatMessage]:
#         """Formats the prompt template into chat messages."""
#         del llm  # unused
#         prompt = self.format(**kwargs)
#         return prompt_to_messages(prompt)

#     def get_template(self, llm: Optional[BaseLLM] = None) -> str:
#         """Get template."""
#         # get kwarg templates
#         kwarg_tmpl_map = {k: "{k}" for k in self.template_vars}

#         # get "raw" template with all the values filled in with {var_name} 
#         template0 = get_formatted_template(self.predict_module, kwarg_tmpl_map)
#         # HACK: there are special 'format' variables of the form ${var_name} that are meant to 
#         # prompt the LLM, but we do NOT want to replace with actual prompt variable values. 
#         # Replace those with double brackets
#         template1 = replace_placeholder(template0)

#         return template1


# # copied from langchain.py
# class Template2Signature(dspy.Signature):
#     """You are a processor for prompts. I will give you a prompt template (Python f-string) for an arbitrary task for other LMs.
# Your job is to prepare three modular pieces: (i) any essential task instructions or guidelines, (ii) a list of variable names for inputs, (iv) the variable name for output."""

#     template = dspy.InputField(format=lambda x: f"```\n\n{x.strip()}\n\n```\n\nLet's now prepare three modular pieces.")
#     essential_instructions = dspy.OutputField()
#     input_keys = dspy.OutputField(desc='comma-separated list of valid variable names')
#     output_key = dspy.OutputField(desc='a valid variable name')


# def build_signature(prompt: PromptTemplate) -> dspy.Signature:
#     """Attempt to build signature from prompt."""
#     # TODO: allow plugging in any llamaindex LLM 
#     gpt4T = dspy.OpenAI(model='gpt-4-1106-preview', max_tokens=4000, model_type='chat')

#     with dspy.context(lm=gpt4T): 
#         parts = dspy.Predict(Template2Signature)(template=prompt.template)

#     inputs = {k.strip(): InputField() for k in parts.input_keys.split(',')}
#     outputs = {k.strip(): OutputField() for k in parts.output_key.split(',')}

#     # dynamically create a pydantic model that subclasses dspy.Signature
#     fields = {
#         k: (str, v) for k, v in {**inputs, **outputs}.items()
#     }
#     signature = make_signature(fields, parts.essential_instructions)
#     return signature
    

# class DSPyComponent(QueryComponent):
#     """DSPy Query Component. 
    
#     Can take in either a predict module directly.
#     TODO: add ability to translate from an existing prompt template / LLM.
    
#     """
#     predict_module: dspy.Predict
#     predict_template: dsp.Template

#     class Config:
#         arbitrary_types_allowed = True

#     def __init__(
#         self,
#         predict_module: dspy.Predict,
#     ) -> None:
#         """Initialize."""
#         return super().__init__(
#             predict_module=predict_module,
#             predict_template=signature_to_template(predict_module.signature),
#         )

#     @classmethod
#     def from_prompt(
#         cls,
#         prompt_template: BasePromptTemplate,
#         # llm: BaseLLM,
#     ) -> "DSPyComponent":
#         """Initialize from prompt template.

#         LLM is a TODO - currently use DSPy LLM classes.
        
#         """
#         signature = build_signature(prompt_template)
#         predict_module = Predict(signature)
#         return cls(predict_module=predict_module)
    
#     def set_callback_manager(self, callback_manager: CallbackManager) -> None:
#         """Set callback manager."""
#         # TODO: implement
#         pass

#     def _validate_component_inputs(self, input: Dict[str, Any]) -> Dict[str, Any]:
#         """Validate component inputs during run_component."""
#         return input

#     def _run_component(self, **kwargs: Any) -> Dict:
#         """Run component."""
#         prediction = self.predict_module(**kwargs)
#         return {
#             k: getattr(prediction, k) for k in self.output_keys.required_keys
#         }

#     async def _arun_component(self, **kwargs: Any) -> Any:
#         """Run component (async)."""
#         # TODO: no async predict module yet
#         return self._run_component(**kwargs)

#     @property
#     def input_keys(self) -> InputKeys:
#         """Input keys."""
#         input_keys = _input_keys_from_template(self.predict_template)
#         return InputKeys.from_keys(input_keys)

#     @property
#     def output_keys(self) -> OutputKeys:
#         """Output keys."""
#         output_keys = _output_keys_from_template(self.predict_template)
#         return OutputKeys.from_keys(output_keys)


# class LlamaIndexModule(dspy.Module):
#     """A module for LlamaIndex.

#     Wraps a QueryPipeline and exposes it as a dspy module for optimization.
    
#     """

#     class Config:
#         arbitrary_types_allowed = True

#     def __init__(self, query_pipeline: QueryPipeline) -> None:
#         """Initialize."""
#         super().__init__()
#         self.query_pipeline = query_pipeline
#         self.predict_modules = []
#         for module in query_pipeline.module_dict.values():
#             if isinstance(module, DSPyComponent):
#                 self.predict_modules.append(module.predict_module)
        

#     def forward(self, **kwargs: Any) -> Dict[str, Any]:
#         """Forward."""
#         output_dict = self.query_pipeline.run(**kwargs, return_values_direct=False)
#         return dspy.Prediction(**output_dict)
    



import functools
import logging
import os
import threading
import uuid
from datetime import datetime
from hashlib import sha256
from typing import Any, Dict, List, Literal, Optional, cast

import litellm
import pydantic
import ujson
from anyio.streams.memory import MemoryObjectSendStream
from asyncer import syncify
from cachetools import LRUCache, cached
from litellm import RetryPolicy

import dspy
from dspy.adapters.base import Adapter
from dspy.clients.openai import OpenAIProvider
from dspy.clients.provider import Provider, TrainingJob
from dspy.clients.utils_finetune import DataFormat, infer_data_format, validate_data_format
from dspy.utils.callback import BaseCallback, with_callbacks

from .base_lm import BaseLM

logger = logging.getLogger(__name__)


class LM(BaseLM):
    """
    A language model supporting chat or text completion requests for use with DSPy modules.
    """

    def __init__(
        self,
        model: str,
        model_type: Literal["chat", "text"] = "chat",
        temperature: float = 0.0,
        max_tokens: int = 1000,
        cache: bool = True,
        callbacks: Optional[List[BaseCallback]] = None,
        num_retries: int = 8,
        provider=None,
        finetuning_model: Optional[str] = None,
        launch_kwargs: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Create a new language model instance for use with DSPy modules and programs.

        Args:
            model: The model to use. This should be a string of the form ``"llm_provider/llm_name"``
                   supported by LiteLLM. For example, ``"openai/gpt-4o"``.
            model_type: The type of the model, either ``"chat"`` or ``"text"``.
            temperature: The sampling temperature to use when generating responses.
            max_tokens: The maximum number of tokens to generate per response.
            cache: Whether to cache the model responses for reuse to improve performance
                   and reduce costs.
            callbacks: A list of callback functions to run before and after each request.
            num_retries: The number of times to retry a request if it fails transiently due to
                         network error, rate limiting, etc. Requests are retried with exponential
                         backoff.
            provider: The provider to use. If not specified, the provider will be inferred from the model.
            finetuning_model: The model to finetune. In some providers, the models available for finetuning is different
                from the models available for inference.
        """
        # Remember to update LM.copy() if you modify the constructor!
        self.model = model
        self.model_type = model_type
        self.cache = cache
        self.provider = provider or self.infer_provider()
        self.callbacks = callbacks or []
        self.kwargs = dict(temperature=temperature, max_tokens=max_tokens, **kwargs)
        self.history = []
        self.callbacks = callbacks or []
        self.num_retries = num_retries
        self.finetuning_model = finetuning_model
        self.launch_kwargs = launch_kwargs

        # TODO(bug): Arbitrary model strings could include the substring "o1-".
        # We should find a more robust way to check for the "o1-" family models.
        if "o1-" in model:
            assert (
                max_tokens >= 5000 and temperature == 1.0
            ), "OpenAI's o1-* models require passing temperature=1.0 and max_tokens >= 5000 to `dspy.LM(...)`"

    @with_callbacks
    def __call__(self, prompt=None, messages=None, **kwargs):
        # Build the request.
        cache = kwargs.pop("cache", self.cache)
        messages = messages or [{"role": "user", "content": prompt}]
        kwargs = {**self.kwargs, **kwargs}

        # Make the request and handle LRU & disk caching.
        if self.model_type == "chat":
            completion = cached_litellm_completion if cache else litellm_completion
        else:
            completion = cached_litellm_text_completion if cache else litellm_text_completion

        response = completion(
            request=dict(model=self.model, messages=messages, **kwargs),
            num_retries=self.num_retries,
        )
        if kwargs.get("logprobs"):
            outputs = [
                {
                    "text": c.message.content if hasattr(c, "message") else c["text"],
                    "logprobs": c.logprobs if hasattr(c, "logprobs") else c["logprobs"],
                }
                for c in response["choices"]
            ]
        else:
            outputs = [c.message.content if hasattr(c, "message") else c["text"] for c in response["choices"]]

        # Logging, with removed api key & where `cost` is None on cache hit.
        kwargs = {k: v for k, v in kwargs.items() if not k.startswith("api_")}
        entry = dict(prompt=prompt, messages=messages, kwargs=kwargs, response=response)
        entry = dict(**entry, outputs=outputs, usage=dict(response["usage"]))
        entry = dict(**entry, cost=response.get("_hidden_params", {}).get("response_cost"))
        entry = dict(
            **entry,
            timestamp=datetime.now().isoformat(),
            uuid=str(uuid.uuid4()),
            model=self.model,
            model_type=self.model_type,
        )
        self.history.append(entry)
        self.update_global_history(entry)

        return outputs

    def launch(self, launch_kwargs: Optional[Dict[str, Any]] = None):
        launch_kwargs = launch_kwargs or self.launch_kwargs
        self.provider.launch(self.model, launch_kwargs)

    def kill(self, launch_kwargs: Optional[Dict[str, Any]] = None):
        launch_kwargs = launch_kwargs or self.launch_kwargs
        self.provider.kill(self.model, launch_kwargs)

    def finetune(
        self,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        data_format: Optional[DataFormat] = None,
    ) -> TrainingJob:
        from dspy import settings as settings

        err = "Fine-tuning is an experimental feature."
        err += " Set `dspy.settings.experimental` to `True` to use it."
        assert settings.experimental, err

        err = f"Provider {self.provider} does not support fine-tuning."
        assert self.provider.finetunable, err

        # Perform data validation before starting the thread to fail early
        train_kwargs = train_kwargs or {}
        if not data_format:
            adapter = self.infer_adapter()
            data_format = infer_data_format(adapter)
        validate_data_format(data=train_data, data_format=data_format)

        # TODO(PR): We can quickly add caching, but doing so requires
        # adding functions that just call other functions as we had in the last
        # iteration, unless people have other ideas.
        def thread_function_wrapper():
            return self._run_finetune_job(job)

        thread = threading.Thread(target=thread_function_wrapper)
        model_to_finetune = self.finetuning_model or self.model
        job = self.provider.TrainingJob(
            thread=thread,
            model=model_to_finetune,
            train_data=train_data,
            train_kwargs=train_kwargs,
            data_format=data_format,
        )
        thread.start()

        return job

    def _run_finetune_job(self, job: TrainingJob):
        # TODO(enhance): We should listen for keyboard interrupts somewhere.
        # Requires TrainingJob.cancel() to be implemented for each provider.
        try:
            model = self.provider.finetune(
                job=job,
                model=job.model,
                train_data=job.train_data,
                train_kwargs=job.train_kwargs,
                data_format=job.data_format,
            )
            lm = self.copy(model=model)
            job.set_result(lm)
        except Exception as err:
            logger.error(err)
            job.set_result(err)

    def infer_provider(self) -> Provider:
        if OpenAIProvider.is_provider_model(self.model):
            return OpenAIProvider()
        # TODO(PR): Keeping this function here will require us to import all
        # providers in this file. Is this okay?
        return Provider()

    def infer_adapter(self) -> Adapter:
        import dspy

        if dspy.settings.adapter:
            return dspy.settings.adapter

        model_type_to_adapter = {
            "chat": dspy.ChatAdapter(),
        }
        model_type = self.model_type
        return model_type_to_adapter[model_type]

    def copy(self, **kwargs):
        """Returns a copy of the language model with possibly updated parameters."""

        import copy

        new_instance = copy.deepcopy(self)
        new_instance.history = []

        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(new_instance, key, value)
            if (key in self.kwargs) or (not hasattr(self, key)):
                new_instance.kwargs[key] = value

        return new_instance


def request_cache(maxsize: Optional[int] = None):
    """
    A threadsafe decorator to create an in-memory LRU cache for LM inference functions that accept
    a dictionary-like LM request. An in-memory cache for LM calls is critical for ensuring
    good performance when optimizing and evaluating DSPy LMs (disk caching alone is too slow).

    Args:
        maxsize: The maximum size of the cache. If unspecified, no max size is enforced (cache is unbounded).

    Returns:
        A decorator that wraps the target function with caching.
    """

    def cache_key(request: Dict[str, Any]) -> str:
        """
        Obtain a unique cache key for the given request dictionary by hashing its JSON
        representation. For request fields having types that are known to be JSON-incompatible,
        convert them to a JSON-serializable format before hashing.

        Note: Values that cannot be converted to JSON should *not* be ignored / discarded, since
        that would potentially lead to cache collisions. For example, consider request A
        containing only JSON-convertible values and request B containing the same JSON-convertible
        values in addition to one unconvertible value. Discarding the unconvertible value would
        lead to a cache collision between requests A and B, even though they are semantically
        different.
        """

        def transform_value(value):
            if isinstance(value, type) and issubclass(value, pydantic.BaseModel):
                return value.schema()
            elif isinstance(value, pydantic.BaseModel):
                return value.dict()
            elif callable(value) and hasattr(value, "__code__") and hasattr(value.__code__, "co_code"):
                return value.__code__.co_code.decode("utf-8")
            else:
                # Note: We don't attempt to compute a hash of the value, since the default
                # implementation of hash() is id(), which may collide if the same memory address
                # is reused for different objects at different times
                return value

        params = {k: transform_value(v) for k, v in request.items()}
        return sha256(ujson.dumps(params, sort_keys=True).encode()).hexdigest()

    def decorator(func):
        @cached(
            # NB: cachetools doesn't support maxsize=None; it recommends using float("inf") instead
            cache=LRUCache(maxsize=maxsize or float("inf")),
            key=lambda key, request, *args, **kwargs: key,
            # Use a lock to ensure thread safety for the cache when DSPy LMs are queried
            # concurrently, e.g. during optimization and evaluation
            lock=threading.RLock(),
        )
        def func_cached(key: str, request: Dict[str, Any], *args, **kwargs):
            return func(request, *args, **kwargs)

        @functools.wraps(func)
        def wrapper(request: dict, *args, **kwargs):
            try:
                key = cache_key(request)
            except Exception:
                # If the cache key cannot be computed (e.g. because it contains a value that cannot
                # be converted to JSON), bypass the cache and call the target function directly
                return func(request, *args, **kwargs)
            return func_cached(key, request, *args, **kwargs)

        return wrapper

    return decorator


@request_cache(maxsize=None)
def cached_litellm_completion(request: Dict[str, Any], num_retries: int):
    return litellm_completion(
        request,
        cache={"no-cache": False, "no-store": False},
        num_retries=num_retries,
    )


def litellm_completion(request: Dict[str, Any], num_retries: int, cache={"no-cache": True, "no-store": True}):
    retry_kwargs = dict(
        retry_policy=_get_litellm_retry_policy(num_retries),
        # In LiteLLM version 1.55.3 (the first version that supports retry_policy as an argument
        # to completion()), the default value of max_retries is non-zero for certain providers, and
        # max_retries is stacked on top of the retry_policy. To avoid this, we set max_retries=0
        max_retries=0,
    )

    stream = dspy.settings.send_stream
    if stream is None:
        return litellm.completion(
            cache=cache,
            **retry_kwargs,
            **request,
        )

    # The stream is already opened, and will be closed by the caller.
    stream = cast(MemoryObjectSendStream, stream)

    @syncify
    async def stream_completion():
        response = await litellm.acompletion(
            cache=cache,
            stream=True,
            **retry_kwargs,
            **request,
        )
        chunks = []
        async for chunk in response:
            chunks.append(chunk)
            await stream.send(chunk)
        return litellm.stream_chunk_builder(chunks)

    return stream_completion()


@request_cache(maxsize=None)
def cached_litellm_text_completion(request: Dict[str, Any], num_retries: int):
    return litellm_text_completion(
        request,
        num_retries=num_retries,
        cache={"no-cache": False, "no-store": False},
    )


def litellm_text_completion(request: Dict[str, Any], num_retries: int, cache={"no-cache": True, "no-store": True}):
    # Extract the provider and model from the model string.
    # TODO: Not all the models are in the format of "provider/model"
    model = request.pop("model").split("/", 1)
    provider, model = model[0] if len(model) > 1 else "openai", model[-1]

    # Use the API key and base from the request, or from the environment.
    api_key = request.pop("api_key", None) or os.getenv(f"{provider}_API_KEY")
    api_base = request.pop("api_base", None) or os.getenv(f"{provider}_API_BASE")

    # Build the prompt from the messages.
    prompt = "\n\n".join([x["content"] for x in request.pop("messages")] + ["BEGIN RESPONSE:"])

    return litellm.text_completion(
        cache=cache,
        model=f"text-completion-openai/{model}",
        api_key=api_key,
        api_base=api_base,
        prompt=prompt,
        retry_policy=_get_litellm_retry_policy(num_retries),
        # In LiteLLM version 1.55.3 (the first version that supports retry_policy as an argument
        # to completion()), the default value of max_retries is non-zero for certain providers, and
        # max_retries is stacked on top of the retry_policy. To avoid this, we set max_retries=0
        max_retries=0,
        **request,
    )


def _get_litellm_retry_policy(num_retries: int) -> RetryPolicy:
    """
    Get a LiteLLM retry policy for retrying requests when transient API errors occur.
    Args:
        num_retries: The number of times to retry a request if it fails transiently due to
                     network error, rate limiting, etc. Requests are retried with exponential
                     backoff.
    Returns:
        A LiteLLM RetryPolicy instance.
    """
    return RetryPolicy(
        TimeoutErrorRetries=num_retries,
        RateLimitErrorRetries=num_retries,
        InternalServerErrorRetries=num_retries,
        ContentPolicyViolationErrorRetries=num_retries,
        # We don't retry on errors that are unlikely to be transient
        # (e.g. bad request, invalid auth credentials)
        BadRequestErrorRetries=0,
        AuthenticationErrorRetries=0,
    )



import inspect
from typing import Any, Callable, Literal, get_origin, get_type_hints

from pydantic import BaseModel

import dspy
from dspy.adapters.json_adapter import get_annotation_name
from dspy.primitives.program import Module
from dspy.signatures.signature import ensure_signature
from dspy.utils.callback import with_callbacks


class Tool:
    def __init__(self, func: Callable, name: str = None, desc: str = None, args: dict[str, Any] = None):
        annotations_func = func if inspect.isfunction(func) or inspect.ismethod(func) else func.__call__
        self.func = func
        self.name = name or getattr(func, "__name__", type(func).__name__)
        self.desc = desc or getattr(func, "__doc__", None) or getattr(annotations_func, "__doc__", "")
        self.args = {
            k: v.schema()
            if isinstance((origin := get_origin(v) or v), type) and issubclass(origin, BaseModel)
            else get_annotation_name(v)
            for k, v in (args or get_type_hints(annotations_func)).items()
            if k != "return"
        }

    @with_callbacks
    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class ReAct(Module):
    def __init__(self, signature, tools: list[Callable], max_iters=5):
        """
        `tools` is either a list of functions, callable classes, or `dspy.Tool` instances.
        """

        self.signature = signature = ensure_signature(signature)
        self.max_iters = max_iters

        tools = [t if isinstance(t, Tool) or hasattr(t, "input_variable") else Tool(t) for t in tools]
        tools = {tool.name: tool for tool in tools}

        inputs = ", ".join([f"`{k}`" for k in signature.input_fields.keys()])
        outputs = ", ".join([f"`{k}`" for k in signature.output_fields.keys()])
        instr = [f"{signature.instructions}\n"] if signature.instructions else []

        instr.extend(
            [
                f"You will be given {inputs} and your goal is to finish with {outputs}.\n",
                "To do this, you will interleave Thought, Tool Name, and Tool Args, and receive a resulting Observation.\n",
                "Thought can reason about the current situation, and Tool Name can be the following types:\n",
            ]
        )

        finish_desc = (
            f"Signals that the final outputs, i.e. {outputs}, are now available and marks the task as complete."
        )
        finish_args = {}  # k: v.annotation for k, v in signature.output_fields.items()}
        tools["finish"] = Tool(func=lambda **kwargs: "Completed.", name="finish", desc=finish_desc, args=finish_args)

        for idx, tool in enumerate(tools.values()):
            args = tool.args if hasattr(tool, "args") else str({tool.input_variable: str})
            desc = (f", whose description is <desc>{tool.desc}</desc>." if tool.desc else ".").replace("\n", "  ")
            desc += f" It takes arguments {args} in JSON format."
            instr.append(f"({idx+1}) {tool.name}{desc}")

        react_signature = (
            dspy.Signature({**signature.input_fields}, "\n".join(instr))
            .append("trajectory", dspy.InputField(), type_=str)
            .append("next_thought", dspy.OutputField(), type_=str)
            .append("next_tool_name", dspy.OutputField(), type_=Literal[tuple(tools.keys())])
            .append("next_tool_args", dspy.OutputField(), type_=dict[str, Any])
        )

        fallback_signature = dspy.Signature(
            {**signature.input_fields, **signature.output_fields}, signature.instructions
        ).append("trajectory", dspy.InputField(), type_=str)

        self.tools = tools
        self.react = dspy.Predict(react_signature)
        self.extract = dspy.ChainOfThought(fallback_signature)

    def forward(self, **input_args):
        def format(trajectory: dict[str, Any], last_iteration: bool):
            adapter = dspy.settings.adapter or dspy.ChatAdapter()
            trajectory_signature = dspy.Signature(f"{', '.join(trajectory.keys())} -> x")
            return adapter.format_fields(trajectory_signature, trajectory, role="user")

        trajectory = {}
        for idx in range(self.max_iters):
            pred = self.react(**input_args, trajectory=format(trajectory, last_iteration=(idx == self.max_iters - 1)))

            trajectory[f"thought_{idx}"] = pred.next_thought
            trajectory[f"tool_name_{idx}"] = pred.next_tool_name
            trajectory[f"tool_args_{idx}"] = pred.next_tool_args

            try:
                trajectory[f"observation_{idx}"] = self.tools[pred.next_tool_name](**pred.next_tool_args)
            except Exception as e:
                trajectory[f"observation_{idx}"] = f"Failed to execute: {e}"

            if pred.next_tool_name == "finish":
                break

        extract = self.extract(**input_args, trajectory=format(trajectory, last_iteration=False))
        return dspy.Prediction(trajectory=trajectory, **extract)


"""
Thoughts and Planned Improvements for dspy.ReAct.

TOPIC 01: How Trajectories are Formatted, or rather when they are formatted.

Right now, both sub-modules are invoked with a `trajectory` argument, which is a string formatted in `forward`. Though
the formatter uses a general adapter.format_fields, the tracing of DSPy only sees the string, not the formatting logic.

What this means is that, in demonstrations, even if the user adjusts the adapter for a fixed program, the demos' format
will not update accordingly, but the inference-time trajectories will.

One way to fix this is to support `format=fn` in the dspy.InputField() for "trajectory" in the signatures. But this
means that care must be taken that the adapter is accessed at `forward` runtime, not signature definition time.

Another potential fix is to more natively support a "variadic" input field, where the input is a list of dictionaries,
or a big dictionary, and have each adatper format it accordingly.

Trajectories also affect meta-programming modules that view the trace later. It's inefficient O(n^2) to view the
trace of every module repeating the prefix.


TOPIC 02: Handling default arguments in the Tool class.


TOPIC 03: Simplifying ReAct's __init__ by moving modular logic to the Tool class.
    * Handling descriptions and casting.
    * Handling exceptions and error messages.
    * More cleanly defining the "finish" tool, perhaps as a runtime-defined function?


TOPIC 04: Default behavior when the trajectory gets too long.


TOPIC 05: Adding more structure around how the instruction is formatted.
    * Concretely, it's now a string, so an optimizer can and does rewrite it freely.
    * An alternative would be to add more structure, such that a certain template is fixed but values are variable?


TOPIC 06: Idiomatically allowing tools that maintain state across iterations, but not across different `forward` calls.
    * So the tool would be newly initialized at the start of each `forward` call, but maintain state across iterations.
    * This is pretty useful for allowing the agent to keep notes or count certain things, etc.

TOPIC 07: Make max_iters a bit more expressive.
    * Allow passing `max_iters` in forward to overwrite the default.
    * Get rid of `last_iteration: bool` in the format function. It's not necessary now.
"""



import random

from datasets import load_dataset

from dspy.datasets.dataset import Dataset


class HotPotQA(Dataset):
    def __init__(
        self,
        *args,
        only_hard_examples=True,
        keep_details="dev_titles",
        unofficial_dev=True,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        assert only_hard_examples, (
            "Care must be taken when adding support for easy examples."
            "Dev must be all hard to match official dev, but training can be flexible."
        )

        hf_official_train = load_dataset("hotpot_qa", "fullwiki", split="train", trust_remote_code=True)
        hf_official_dev = load_dataset("hotpot_qa", "fullwiki", split="validation", trust_remote_code=True)

        official_train = []
        for raw_example in hf_official_train:
            if raw_example["level"] == "hard":
                if keep_details is True:
                    keys = ["id", "question", "answer", "type", "supporting_facts", "context"]
                elif keep_details == "dev_titles":
                    keys = ["question", "answer", "supporting_facts"]
                else:
                    keys = ["question", "answer"]

                example = {k: raw_example[k] for k in keys}

                if "supporting_facts" in example:
                    example["gold_titles"] = set(example["supporting_facts"]["title"])
                    del example["supporting_facts"]

                official_train.append(example)

        rng = random.Random(0)
        rng.shuffle(official_train)

        self._train = official_train[: len(official_train) * 75 // 100]

        if unofficial_dev:
            self._dev = official_train[len(official_train) * 75 // 100 :]
        else:
            self._dev = None

        for example in self._train:
            if keep_details == "dev_titles":
                del example["gold_titles"]

        test = []
        for raw_example in hf_official_dev:
            assert raw_example["level"] == "hard"
            example = {k: raw_example[k] for k in ["id", "question", "answer", "type", "supporting_facts"]}
            if "supporting_facts" in example:
                example["gold_titles"] = set(example["supporting_facts"]["title"])
                del example["supporting_facts"]
            test.append(example)

        self._test = test


if __name__ == "__main__":
    from dspy.dsp.utils import dotdict

    data_args = dotdict(train_seed=1, train_size=16, eval_seed=2023, dev_size=200 * 5, test_size=0)
    dataset = HotPotQA(**data_args)

    print(dataset)
    print(dataset.train[0].question)
    print(dataset.train[15].question)

    print(len(dataset.train), len(dataset.dev), len(dataset.test))

    print(dataset.dev[0].question)
    print(dataset.dev[340].question)
    print(dataset.dev[937].question)

"""
What was the population of the city where Woodward Avenue ends in 2010?
Where did the star , who is also an executive producer, of the Mick begin her carrer? 
16 1000 0
Both London and German have seen attacks during war, there was one specific type of attack that Germany called the blitz, what did London call a similar attack?
Pre-Madonna was a collection of demos by the singer who was a leading presence during the emergence of what network?
Alan Mills composed the classic folk song that tells the story of what? 
"""



from dspy.dsp.utils import EM, normalize_text

from dspy.evaluate.metrics import answer_exact_match, answer_passage_match
from dspy.evaluate.evaluate import Evaluate
from dspy.evaluate.auto_evaluation import SemanticF1, CompleteAndGrounded



from dspy.teleprompt.avatar_optimizer import AvatarOptimizer
from dspy.teleprompt.bettertogether import BetterTogether
from dspy.teleprompt.bootstrap import BootstrapFewShot
from dspy.teleprompt.bootstrap_finetune import BootstrapFinetune
from dspy.teleprompt.copro_optimizer import COPRO
from dspy.teleprompt.ensemble import Ensemble
from dspy.teleprompt.knn_fewshot import KNNFewShot

# from .mipro_optimizer import MIPRO
from dspy.teleprompt.mipro_optimizer_v2 import MIPROv2
from dspy.teleprompt.random_search import BootstrapFewShotWithRandomSearch

# from .signature_opt import SignatureOptimizer
# from .signature_opt_bayesian import BayesianSignatureOptimizer
from dspy.teleprompt.teleprompt import Teleprompter
from dspy.teleprompt.teleprompt_optuna import BootstrapFewShotWithOptuna
from dspy.teleprompt.vanilla import LabeledFewShot

__all__ = [
    "AvatarOptimizer",
    "BetterTogether",
    "BootstrapFewShot",
    "BootstrapFinetune",
    "COPRO",
    "Ensemble",
    "KNNFewShot",
    "MIPROv2",
    "BootstrapFewShotWithRandomSearch",
    "BootstrapFewShotWithOptuna",
    "LabeledFewShot",
]



from typing import Literal

from pydantic.fields import FieldInfo


def get_dspy_field_type(field: FieldInfo) -> Literal["input", "output"]:
    field_type = field.json_schema_extra.get("__dspy_field_type")
    if field_type is None:
        raise ValueError(f"Field {field} does not have a __dspy_field_type")
    return field_type



from abc import abstractmethod
from concurrent.futures import Future
from threading import Thread
from typing import Any, Dict, List, Optional, Union

from dspy.clients.utils_finetune import DataFormat


class TrainingJob(Future):
    def __init__(
        self,
        thread: Optional[Thread] = None,
        model: Optional[str] = None,
        train_data: Optional[List[Dict[str, Any]]] = None,
        train_kwargs: Optional[Dict[str, Any]] = None,
        data_format: Optional[DataFormat] = None,
    ):
        self.thread = thread
        self.model = model
        self.train_data = train_data
        self.train_kwargs = train_kwargs or {}
        self.data_format = data_format
        super().__init__()

    # Subclasses should override the cancel method to cancel the job; then call
    # the super's cancel method so that the future can be cancelled.
    def cancel(self):
        super().cancel()

    @abstractmethod
    def status(self):
        raise NotImplementedError


class Provider:
    def __init__(self):
        self.finetunable = False
        self.TrainingJob = TrainingJob

    @staticmethod
    def is_provider_model(model: str) -> bool:
        # Subclasses should actually check whether a model is supported if they
        # want to have the model provider auto-discovered.
        return False

    @staticmethod
    def launch(model: str, launch_kwargs: Optional[Dict[str, Any]] = None):
        pass

    @staticmethod
    def kill(model: str, kill_kwargs: Optional[Dict[str, Any]] = None):
        pass

    @staticmethod
    def finetune(
        job: TrainingJob,
        model: str,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        data_format: Optional[Union[DataFormat, str]] = None,
    ) -> str:
        raise NotImplementedError



"""
Retriever model for Milvus or Zilliz Cloud
"""

from typing import Callable, List, Optional, Union

import dspy
from dspy.dsp.utils import dotdict

try:
    from pymilvus import MilvusClient
except ImportError:
    raise ImportError(
        "The pymilvus library is required to use MilvusRM. Install it with `pip install dspy-ai[milvus]`",
    )


def openai_embedding_function(texts: List[str]):
    from openai import OpenAI

    client = OpenAI()
    response = client.embeddings.create(
        input=texts,
        model="text-embedding-3-small",
    )
    return [x.embedding for x in response.data]


class MilvusRM(dspy.Retrieve):
    """
    A retrieval module that uses Milvus to return passages for a given query.

    Assumes that a Milvus collection has been created and populated with the following field:
        - text: The text of the passage

    Args:
        collection_name (str): The name of the Milvus collection to query against.
        uri (str, optional): The Milvus connection uri. Defaults to "http://localhost:19530".
        token (str, optional): The Milvus connection token. Defaults to None.
        db_name (str, optional): The Milvus database name. Defaults to "default".
        embedding_function (callable, optional): The function to convert a list of text to embeddings. 
            The embedding function should take a list of text strings as input and output a list of embeddings.
            Defaults to None. By default, it will get OpenAI client by the environment variable OPENAI_API_KEY 
            and use OpenAI's embedding model "text-embedding-3-small" with the default dimension.
        k (int, optional): The number of top passages to retrieve. Defaults to 3.

    Returns:
        dspy.Prediction: An object containing the retrieved passages.

    Examples:
        Below is a code snippet that shows how to use this as the default retriever:
        ```python
        llm = dspy.OpenAI(model="gpt-3.5-turbo")
        retriever_model = MilvusRM(
            collection_name="<YOUR_COLLECTION_NAME>",
            uri="<YOUR_MILVUS_URI>",
            token="<YOUR_MILVUS_TOKEN>"
            )
        dspy.settings.configure(lm=llm, rm=retriever_model)
        ```

        Below is a code snippet that shows how to use this in the forward() function of a module
        ```python
        self.retrieve = MilvusRM(k=num_passages)
        ```
    """

    def __init__(
        self,
        collection_name: str,
        uri: Optional[str] = "http://localhost:19530",
        token: Optional[str] = None,
        db_name: Optional[str] = "default",
        embedding_function: Optional[Callable] = None,
        k: int = 3,
    ):
        self.milvus_client = MilvusClient(uri=uri, token=token, db_name=db_name)

        # check collection existence
        if collection_name not in self.milvus_client.list_collections():
            raise AttributeError(f"Milvus collection not found: {collection_name}")
        self.collection_name = collection_name

        self.embedding_function = embedding_function or openai_embedding_function
        self.k = k

    def forward(self, query_or_queries: Union[str, List[str]], k: Optional[int] = None) -> dspy.Prediction:
        if isinstance(query_or_queries, str):
            query_or_queries = [query_or_queries]
        query_embeddings = self.embedding_function(query_or_queries)

        k = k or self.k

        milvus_res = self.milvus_client.search(
            collection_name=self.collection_name,
            data=query_embeddings,
            output_fields=["text"],
            limit=k,
        )
        passage_scores = {}
        for res in milvus_res:
            for r in res:
                passage_scores[r["entity"]["text"]] = r["distance"]
        sorted_passages = sorted(
            passage_scores.items(), key=lambda x: x[1], reverse=True,
        )[:k]
        return [dotdict({"long_text": passage}) for passage, _ in sorted_passages]


from dspy.primitives import assertions
from dspy.primitives.example import Example
from dspy.primitives.module import BaseModule
from dspy.primitives.prediction import Prediction, Completions
from dspy.primitives.program import Program, Module
from dspy.primitives.python_interpreter import PythonInterpreter, TextPrompt, CodePrompt


__all__ = [
    "assertions",
    "Example",
    "BaseModule",
    "Prediction",
    "Completions",
    "Program",
    "Module",
    "PythonInterpreter",
    "TextPrompt",
    "CodePrompt",
]
